#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import argparse

from typing import Text
from ai_flow.endpoint.server.server import AIFlowServer
from ai_flow.endpoint.server.server_config import AIFlowServerConfig
from ai_flow.client.ai_flow_client import get_ai_flow_client
import logging

_SQLITE_DB_FILE = 'aiflow.db'
_SQLITE_DB_URI = '%s%s' % ('sqlite:///', _SQLITE_DB_FILE)
_MYSQL_DB_URI = 'mysql+pymysql://root:aliyunmysql@localhost:3306/aiflow'
_PORT = '50051'
GLOBAL_MASTER_CONFIG = {}


class AIFlowServerRunner(object):
    """
    AI flow server runner. This class is the runner class for the AIFlowServer. It parse the server configuration and
    manage the live cycle of the AIFlowServer.
    """

    def __init__(self, config_file: Text = None, enable_ha=False, server_uri: str = None, ttl_ms=10000) -> None:
        """
        Set the server attribute according to the server config file.

        :param config_file: server configuration file.
        """
        super().__init__()
        self.config_file = config_file
        self.server = None
        self.server_config = AIFlowServerConfig()
        self.enable_ha = enable_ha
        self.server_uri = server_uri
        self.ttl_ms = ttl_ms

    def start(self,
              is_block=False) -> None:
        """
        Start the AI flow runner.

        :param is_block: AI flow runner will run non-stop if True.
        """
        if self.config_file is not None:
            self.server_config.load_from_file(self.config_file)
        else:
            self.server_config.set_server_port(str(_PORT))
        global GLOBAL_MASTER_CONFIG
        GLOBAL_MASTER_CONFIG = self.server_config
        logging.info("AI Flow Master Config {}".format(GLOBAL_MASTER_CONFIG))
        self.server = AIFlowServer(
            store_uri=self.server_config.get_db_uri(),
            port=str(self.server_config.get_server_port()),
            start_default_notification=self.server_config.start_default_notification(),
            notification_uri=self.server_config.get_notification_uri(),
            start_meta_service=self.server_config.start_meta_service(),
            start_model_center_service=self.server_config.start_model_center_service(),
            start_metric_service=self.server_config.start_metric_service(),
            start_scheduler_service=self.server_config.start_scheduler_service(),
            scheduler_service_config=self.server_config.get_scheduler_config(),
            enabled_ha=self.server_config.get_enable_ha(),
            ha_server_uri=self.server_config.get_server_ip() + ":" + str(self.server_config.get_server_port()),
            ttl_ms=self.server_config.get_ha_ttl_ms())
        self.server.run(is_block=is_block)

    def stop(self, clear_sql_lite_db_file=True) -> None:
        """
        Stop the AI flow runner.

        :param clear_sql_lite_db_file: If True, the sqlite database files will be deleted When the server stops working.
        """
        self.server.stop(clear_sql_lite_db_file)

    def _clear_db(self):
        self.server._clear_db()


def set_master_config():
    code, config, message = get_ai_flow_client().get_master_config()
    for k, v in config.items():
        GLOBAL_MASTER_CONFIG[k] = v


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', required=True, help='master config file')
    args = parser.parse_args()
    logging.info(args.config)
    config_file = args.config
    master = AIFlowServerRunner(
        config_file=config_file)
    master.start(is_block=True)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Optional

from flask import current_app, g
from sqlalchemy import and_
from sqlalchemy.orm.session import Session

from airflow.api_connexion import security
from airflow.api_connexion.exceptions import NotFound
from airflow.api_connexion.parameters import check_limit, format_parameters
from airflow.api_connexion.schemas.xcom_schema import (
    XComCollection,
    XComCollectionItemSchema,
    XComCollectionSchema,
    xcom_collection_item_schema,
    xcom_collection_schema,
)
from airflow.models import DagRun as DR, XCom
from airflow.security import permissions
from airflow.utils.session import provide_session


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG_RUN),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_EVENT),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_TASK_EXECUTION),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_TASK_INSTANCE),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_XCOM),
    ]
)
@format_parameters({'limit': check_limit})
@provide_session
def get_xcom_entries(
    dag_id: str,
    dag_run_id: str,
    task_id: str,
    session: Session,
    limit: Optional[int],
    offset: Optional[int] = None,
) -> XComCollectionSchema:
    """Get all XCom values"""
    query = session.query(XCom)
    if dag_id == '~':
        appbuilder = current_app.appbuilder
        readable_dag_ids = appbuilder.sm.get_readable_dag_ids(g.user)
        query = query.filter(XCom.dag_id.in_(readable_dag_ids))
        query = query.join(DR, and_(XCom.dag_id == DR.dag_id, XCom.execution_date == DR.execution_date))
    else:
        query = query.filter(XCom.dag_id == dag_id)
        query = query.join(DR, and_(XCom.dag_id == DR.dag_id, XCom.execution_date == DR.execution_date))

    if task_id != '~':
        query = query.filter(XCom.task_id == task_id)
    if dag_run_id != '~':
        query = query.filter(DR.run_id == dag_run_id)
    query = query.order_by(XCom.execution_date, XCom.task_id, XCom.dag_id, XCom.key)
    total_entries = query.count()
    query = query.offset(offset).limit(limit)
    return xcom_collection_schema.dump(XComCollection(xcom_entries=query.all(), total_entries=total_entries))


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG_RUN),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_EVENT),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_TASK_EXECUTION),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_TASK_INSTANCE),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_XCOM),
    ]
)
@provide_session
def get_xcom_entry(
    dag_id: str, task_id: str, dag_run_id: str, xcom_key: str, session: Session
) -> XComCollectionItemSchema:
    """Get an XCom entry"""
    query = session.query(XCom).filter(XCom.dag_id == dag_id, XCom.task_id == task_id, XCom.key == xcom_key)
    query = query.join(DR, and_(XCom.dag_id == DR.dag_id, XCom.execution_date == DR.execution_date))
    query = query.filter(DR.run_id == dag_run_id)

    query_object = query.one_or_none()
    if not query_object:
        raise NotFound("XCom entry not found")
    return xcom_collection_item_schema.dump(query_object)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
from json import JSONDecodeError

from wtforms.validators import EqualTo, ValidationError


class GreaterEqualThan(EqualTo):
    """Compares the values of two fields.

    :param fieldname:
        The name of the other field to compare to.
    :param message:
        Error message to raise in case of a validation error. Can be
        interpolated with `%(other_label)s` and `%(other_name)s` to provide a
        more helpful error.
    """

    def __call__(self, form, field):
        try:
            other = form[self.fieldname]
        except KeyError:
            raise ValidationError(field.gettext("Invalid field name '%s'." % self.fieldname))

        if field.data is None or other.data is None:
            return

        if field.data < other.data:
            message_args = {
                'other_label': hasattr(other, 'label') and other.label.text or self.fieldname,
                'other_name': self.fieldname,
            }
            message = self.message
            if message is None:
                message = field.gettext(
                    'Field must be greater than or equal to %(other_label)s.' % message_args
                )
            else:
                message = message % message_args

            raise ValidationError(message)


class ValidJson:
    """Validates data is valid JSON.

    :param message:
        Error message to raise in case of a validation error.
    """

    def __init__(self, message=None):
        self.message = message

    def __call__(self, form, field):
        if field.data:
            try:
                json.loads(field.data)
            except JSONDecodeError as ex:
                message = self.message or f'JSON Validation Error: {ex}'
                raise ValidationError(message=field.gettext(message.format(field.data)))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Set


class TriggerRule:
    """Class with task's trigger rules."""

    ALL_SUCCESS = 'all_success'
    ALL_FAILED = 'all_failed'
    ALL_DONE = 'all_done'
    ONE_SUCCESS = 'one_success'
    ONE_FAILED = 'one_failed'
    NONE_FAILED = 'none_failed'
    NONE_FAILED_OR_SKIPPED = 'none_failed_or_skipped'
    NONE_SKIPPED = 'none_skipped'
    DUMMY = 'dummy'

    _ALL_TRIGGER_RULES: Set[str] = set()

    @classmethod
    def is_valid(cls, trigger_rule):
        """Validates a trigger rule."""
        return trigger_rule in cls.all_triggers()

    @classmethod
    def all_triggers(cls):
        """Returns all trigger rules."""
        if not cls._ALL_TRIGGER_RULES:
            cls._ALL_TRIGGER_RULES = {
                getattr(cls, attr)
                for attr in dir(cls)
                if not attr.startswith("_") and not callable(getattr(cls, attr))
            }
        return cls._ALL_TRIGGER_RULES



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import tempfile
import uuid
import zipfile
import oss2
import warnings
import tensorflow
import pickle

from distutils.version import LooseVersion
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Text, Dict

from ai_flow.endpoint.server.exception import AIFlowException
from ai_flow.util.file_util.zip_file_util import make_dir_zipfile, make_file_zipfile


class ModelManager(ABC):

    @abstractmethod
    def download_model(self, remote_path, local_path=None):
        pass

    @abstractmethod
    def save_model(self, local_path, remote_path=None):
        pass


class OSSModelManager(ModelManager):

    def __init__(self, config: Dict[Text, Text]):
        ack_id = config.get('blob_server.access_key_id', None)
        ack_secret = config.get('blob_server.access_key_secret', None)
        endpoint = config.get('blob_server.endpoint', None)
        bucket_name = config.get('blob_server.bucket', None)
        auth = oss2.Auth(ack_id, ack_secret)
        self.bucket = oss2.Bucket(auth, endpoint, bucket_name)

    def download_model(self, remote_path, local_path=None):
        """
        download an uploaded model in Aliyun OSS.
        :param remote_path: the object key of the file in oss.
        :param local_path: the user specified local path of the downloaded file. the local_path must be a directory
        :return: the downloaded file path.
        """
        if local_path is not None:
            local_file_dir = local_path
        else:
            local_file_dir = 'tmp_model-' + str(uuid.uuid1())
        tmp_dir = Path(tempfile.gettempdir())
        tmp_zip_file = str(tmp_dir / local_file_dir) + '.zip'
        oss_object_key = remote_path
        self.bucket.get_object_to_file(key=oss_object_key, filename=tmp_zip_file)
        with zipfile.ZipFile(tmp_zip_file, 'r') as zip_ref:
            top_name = zip_ref.namelist()[0]
            extract_path = str(tmp_dir / local_file_dir)
            zip_ref.extractall(extract_path)
        downloaded_local_path = Path(extract_path) / top_name
        return str(downloaded_local_path)

    def save_model(self, local_path, remote_path=None) -> Text:
        """
        save a local export model to remote storage(Aliyun OSS)
        :param local_path: the local path of the model, the oss storage only supports binary files, thus we have to
        make the local path file a zip.
        :param remote_path: the object_key of the uploaded file in oss with the pattern like abc/efg/123.jpg
        :return: the object_key of uploaded file.
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            local_path = Path(local_path)
            zip_file_name = os.path.splitext(local_path.name)[0] + '.zip'
            temp_dir_path = Path(temp_dir)
            zip_file_path = temp_dir_path / zip_file_name
            if local_path.is_dir():
                make_dir_zipfile(str(local_path), zip_file_path)
            else:
                make_file_zipfile(str(local_path), zip_file_path)
            if remote_path is None:
                object_key = 'ai-flow-model-manager/' + zip_file_name
            else:
                object_key = remote_path
            self.bucket.put_object_from_file(key=object_key, filename=str(zip_file_path))
        return object_key


def _load_model_from_local_file(path):
    """Load a scikit-learn model saved as an aiflow artifact on the local file system."""
    with open(path, "rb") as f:
        return pickle.load(f)


def load_scikit_learn_model(model_uri):
    """
    Load a scikit-learn model from a local file.

    :param model_uri: The location, in URI format, of the aiflow model, for example:

                      - ``/Users/aiflow/path/to/local/model``
                      - ``relative/path/to/local/model``
                      - ``s3://my_bucket/path/to/model``
                      - ``models:/<model_name>/<model_version>``
                      - ``models:/<model_name>/<stage>``

    :return: A scikit-learn model.
    """
    return _load_model_from_local_file(path=model_uri)


def _load_tensorflow_saved_model(model_uri, meta_graph_tags, signature_def_map_key, tf_session):
    """
    Load a specified TensorFlow model consisting of a TensorFlow metagraph and signature definition
    from a serialized TensorFlow ``SavedModel`` collection.

    :param model_uri: The local filesystem path or run-relative artifact path to the model.
    :param meta_graph_tags: A list of tags identifying the model's metagraph within the
                               serialized ``SavedModel`` object. For more information, see the
                               ``tags`` parameter of the `tf.saved_model.builder.SavedModelBuilder
                               method <https://www.tensorflow.org/api_docs/python/tf/saved_model/
                               builder/SavedModelBuilder#add_meta_graph>`_.
    :param signature_def_map_key: A string identifying the input/output signature associated with the
                                 model. This is a key within the serialized ``SavedModel``'s
                                 signature definition mapping. For more information, see the
                                 ``signature_def_map`` parameter of the
                                 ``tf.saved_model.builder.SavedModelBuilder`` method.
    :param tf_session: The TensorFlow session in which to load the metagraph.
                    Required in TensorFlow versions < 2.0.0. Unused in TensorFlow versions >= 2.0.0
    :return: For TensorFlow versions < 2.0.0:
             A TensorFlow signature definition of type:
             ``tensorflow.core.protobuf.meta_graph_pb2.SignatureDef``. This defines input and
             output tensors within the specified metagraph for inference.
             For TensorFlow versions >= 2.0.0:
             A callable graph (tensorflow.function) that takes inputs and returns inferences.
    """
    if LooseVersion(tensorflow.__version__) < LooseVersion('2.0.0'):
        loaded = tensorflow.saved_model.loader.load(
            sess=tf_session,
            tags=meta_graph_tags,
            export_dir=model_uri)
        loaded_sig = loaded.signature_def
    else:
        loaded = tensorflow.saved_model.load(  # pylint: disable=no-value-for-parameter
            tags=meta_graph_tags,
            export_dir=model_uri)
        loaded_sig = loaded.signatures
    if signature_def_map_key not in loaded_sig:
        raise AIFlowException("Could not find signature def key %s. Available keys are: %s"
                              % (signature_def_map_key, list(loaded_sig.keys())))
    return loaded_sig[signature_def_map_key]


def load_tensorflow_saved_model(model_uri, meta_graph_tags, signature_def_map_key, tf_session=None):
    """
    Load a tensorflow model from a specific path.

    *With TensorFlow version <2.0.0, this method must be called within a TensorFlow graph context.*

    :param model_uri: The location, in URI format, of the tensorflow model. For example:

                      - ``/Users/me/path/to/local/model``
                      - ``relative/path/to/local/model``

    :param meta_graph_tags: A list of tags identifying the model's metagraph within the
                               serialized ``SavedModel`` object. For more information, see the
                               ``tags`` parameter of the `tf.saved_model.builder.SavedModelBuilder
                               method <https://www.tensorflow.org/api_docs/python/tf/saved_model/
                               builder/SavedModelBuilder#add_meta_graph>`_.

    :param signature_def_map_key: A string identifying the input/output signature associated with the
                                 model. This is a key within the serialized ``SavedModel``'s
                                 signature definition mapping. For more information, see the
                                 ``signature_def_map`` parameter of the
                                 ``tf.saved_model.builder.SavedModelBuilder`` method.

    :param tf_session: The TensorFlow session in which to load the model. If using TensorFlow
                    version >= 2.0.0, this argument is ignored. If using TensorFlow <2.0.0, if no
                    session is passed to this function, aiflow will attempt to load the model using
                    the default TensorFlow session.  If no default session is available, then the
                    function raises an exception.
    :return: For TensorFlow < 2.0.0, a TensorFlow signature definition of type:
             ``tensorflow.core.protobuf.meta_graph_pb2.SignatureDef``. This defines the input and
             output tensors for model inference.
             For TensorFlow >= 2.0.0, A callable graph (tf.function) that takes inputs and
             returns inferences.
    """
    if LooseVersion(tensorflow.__version__) < LooseVersion('2.0.0'):
        if not tf_session:
            tf_session = tensorflow.get_default_session()
            if not tf_session:
                raise AIFlowException("No TensorFlow session found while calling load_model()." +
                                      "You can set the default Tensorflow session before calling" +
                                      " load_model via `session.as_default()`, or directly pass " +
                                      "a session in which to load the model via the tf_sess " +
                                      "argument.")

    else:
        if tf_session:
            warnings.warn("A TensorFlow session was passed into load_model, but the " +
                          "currently used version is TF 2.0 where sessions are deprecated. " +
                          "The tf_sess argument will be ignored.", FutureWarning)
    return _load_tensorflow_saved_model(model_uri=model_uri,
                                        meta_graph_tags=meta_graph_tags,
                                        signature_def_map_key=signature_def_map_key,
                                        tf_session=tf_session)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import json
import subprocess
import time
import unittest
import requests

HEADER = 'http://localhost:8080'

"""
start server service first
the default server port is 50051 and the proxy port is 8080
make sure no port conflicts
"""


class Test_Rest_Api(unittest.TestCase):
    def setUp(self) -> None:
        subprocess.Popen("cd sbin && bash start_service.sh", shell=True)
        # wait the server and proxy to start
        time.sleep(3)

    def tearDown(self) -> None:
        subprocess.Popen("cd sbin && bash stop_service.sh", shell=True)

    def test_save_example_get_id_name(self):
        example = {
            "example": {
                "name": "example",
                "properties": {
                    "a": "b"
                },
                "description": "wow!",
                "supportType": "EXAMPLE_BOTH",
                "schema": {
                    "name_list": ["a", "b", "c"],
                    "type_list": ["Int32", "Int64", "String"]
                },
                "format": "csv"
            }
        }
        data = json.dumps(example)
        response = requests.post(HEADER + '/aiflow/metadata_store/dataset/save', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(int(result_data.get('uuid')), 1)

        data = '{"id":1}'
        response = requests.post(HEADER + '/aiflow/metadata_store/dataset/get/id', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('name'), 'example')

        data = '{"name":"example"}'
        response = requests.post(HEADER + '/aiflow/metadata_store/dataset/get/name', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('name'), 'example')

        examples = {
            "examples": [{
                "name": "example1"
            }, {
                "name": "example2"
            }]
        }
        data = json.dumps(examples)
        response = requests.post(HEADER + '/aiflow/metadata_store/dataset/save', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(len(result_data.get('examples')), 2)

        data = '{"pageSize":2,"offset":1}'
        response = requests.post(HEADER + '/aiflow/metadata_store/dataset/list', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(2, len(result_data.get('examples')))

    def test_project(self):
        project = {
            "project": {"name": "project"}
        }
        data = json.dumps(project)
        response = requests.post(HEADER + '/aiflow/metadata_store/project/save', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(int(result_data.get('uuid')), 1)

        data = '{"id":1}'
        response = requests.post(HEADER + '/aiflow/metadata_store/project/get/id', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('name'), 'project')

        data = '{"name":"project"}'
        response = requests.post(HEADER + '/aiflow/metadata_store/project/get/name', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('name'), 'project')

        data = '{"pageSize":2,"offset":0}'
        response = requests.post(HEADER + '/aiflow/metadata_store/project/list', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(1, len(result_data.get('projects')))

    def test_model(self):
        model = {
            "model": {"name": "model",
                      "projectId": 1}
        }
        data = json.dumps(model)
        response = requests.post(HEADER + '/aiflow/metadata_store/model/save', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(int(result_data.get('uuid')), 1)

        data = '{"id":1}'
        response = requests.post(HEADER + '/aiflow/metadata_store/model/get/id', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('name'), 'model')

        data = '{"name":"model"}'
        response = requests.post(HEADER + '/aiflow/metadata_store/model/get/name', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('name'), 'model')

        data = '{"pageSize":2,"offset":0}'
        response = requests.post(HEADER + '/aiflow/metadata_store/model/list', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(1, len(result_data.get('models')))

    def test_model_version(self):
        model_version = {
            "modelVersion": {"version": "model version",
                             "modelId": 1,
                             "workflowExecutionId": 1}
        }
        data = json.dumps(model_version)
        response = requests.post(HEADER + '/aiflow/metadata_store/modelVersion/save', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('version'), 'model version')

        data = '{"name":"model version"}'
        response = requests.post(HEADER + '/aiflow/metadata_store/modelVersion/get/version', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('version'), 'model version')

        version = '{"name":"model version"}'
        requests.post(HEADER + '/aiflow/metadata_store/modelVersion/delete/version', data=version, verify=False)

        data = '{"modelId":1, "pageSize":2,"offset":0}'
        response = requests.post(HEADER + '/aiflow/metadata_store/modelVersion/list', data=data, verify=False)
        return_code = json.loads(response.text).get('return_code')
        self.assertEqual(return_code, '2002')

    def test_workflow_execution(self):
        workflow_execution = {
            "workflowExecution": {"name": "execution",
                                  "projectId": 1}
        }
        data = json.dumps(workflow_execution)
        response = requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/save', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(int(result_data.get('uuid')), 1)

        data = '{"id":1}'
        response = requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/get/id', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('name'), 'execution')

        data = '{"name":"execution"}'
        response = requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/get/name', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('name'), 'execution')

        data = '{"pageSize":2,"offset":0}'
        response = requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/list', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(1, len(result_data.get('workflow_executions')))

    def test_job(self):
        job = {
            "job": {"name": "job",
                    "workflowExecutionId": 1}
        }
        data = json.dumps(job)
        response = requests.post(HEADER + '/aiflow/metadata_store/job/save', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(int(result_data.get('uuid')), 1)

        data = '{"id":1}'
        response = requests.post(HEADER + '/aiflow/metadata_store/job/get/id', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('name'), 'job')

        data = '{"state": "FINISHED", "name":"job"}'
        response = requests.post(HEADER + '/aiflow/metadata_store/job/update', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data, 1)

        data = '{"name":"job"}'
        response = requests.post(HEADER + '/aiflow/metadata_store/job/get/name', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        self.assertEqual(result_data.get('name'), 'job')
        self.assertEqual(result_data.get('state'), 'FINISHED')

        data = '{"id":1}'
        requests.post(HEADER + '/aiflow/metadata_store/job/delete/id', data=data, verify=False)

        job_1 = {
            "job": {"name": "job1",
                    "workflowExecutionId": 1}
        }
        data = json.dumps(job_1)
        requests.post(HEADER + '/aiflow/metadata_store/job/save', data=data, verify=False)

        data = '{"name":"job1"}'
        requests.post(HEADER + '/aiflow/metadata_store/job/delete/name', data=data, verify=False)

        data = '{"pageSize":2,"offset":0}'
        response = requests.post(HEADER + '/aiflow/metadata_store/job/list', data=data, verify=False)
        return_code = json.loads(response.text).get('return_code')
        self.assertEqual(return_code, '2002')


class Test_delete_rest_api(unittest.TestCase):

    def setUp(self) -> None:
        subprocess.Popen("cd sbin && bash start_service.sh", shell=True)
        # wait the server and proxy to start
        time.sleep(3)

    def tearDown(self) -> None:
        subprocess.Popen("cd sbin && bash stop_service.sh", shell=True)

    def test_delete_example_by_id(self):
        example = {
            "example": {
                "name": "example",
                "properties": {
                    "a": "b"
                },
                "supportType": "EXAMPLE_BOTH",
                "format": "csv"
            }
        }
        data = json.dumps(example)
        response = requests.post(HEADER + '/aiflow/metadata_store/dataset/save', data=data, verify=False)
        result_data = json.loads(json.loads(response.text).get('data'))
        uuid = int(result_data.get('uuid'))

        data = '{"id":' + str(uuid) + '}'
        requests.post(HEADER + '/aiflow/metadata_store/dataset/delete/id', data=data, verify=False)

        data = '{"id":1}'
        response = requests.post(HEADER + '/aiflow/metadata_store/dataset/get/id', data=data, verify=False)
        return_code = json.loads(response.text).get('return_code')
        self.assertEqual(return_code, '2002')

    def test_delete_example_by_name(self):
        example = {
            "example": {
                "name": "example1",
                "properties": {
                    "a": "b"
                },
                "supportType": "EXAMPLE_BOTH",
                "format": "csv"
            }
        }
        data = json.dumps(example)
        requests.post(HEADER + '/aiflow/metadata_store/dataset/save', data=data, verify=False)

        data = '{"name":"example1"}'
        requests.post(HEADER + '/aiflow/metadata_store/dataset/delete/name', data=data, verify=False)

        response = requests.post(HEADER + '/aiflow/metadata_store/dataset/get/name', data=data, verify=False)
        return_code = json.loads(response.text).get('return_code')
        self.assertEqual(return_code, '2002')

    def test_delete_project_by_id(self):
        project = {
            "project": {"name": "project"}
        }
        model = {
            "model": {"name": "model",
                      "projectId": 1}
        }
        workflow_execution = {
            "workflowExecution": {"name": "execution",
                                  "projectId": 1}
        }
        model_version = {
            "modelVersion": {"version": "model version",
                             "modelId": 1,
                             "workflowExecutionId": 1}
        }
        job = {
            "job": {"name": "job",
                    "workflowExecutionId": 1}
        }
        requests.post(HEADER + '/aiflow/metadata_store/project/save', data=json.dumps(project), verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/model/save', data=json.dumps(model), verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/save', data=json.dumps(workflow_execution),
                      verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/modelVersion/save', data=json.dumps(model_version), verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/job/save', data=json.dumps(job), verify=False)

        data = '{"id":1}'
        requests.post(HEADER + '/aiflow/metadata_store/project/delete/id', data=data, verify=False)
        self.assertEqual('2002',
                         json.loads(requests.post(HEADER + '/aiflow/metadata_store/project/get/id', data=data,
                                                  verify=False).text).get('return_code'))
        self.assertEqual('2002',
                         json.loads(requests.post(HEADER + '/aiflow/metadata_store/model/get/id', data=data,
                                                  verify=False).text).get('return_code'))
        self.assertEqual('2002',
                         json.loads(requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/get/id',
                                                  data=data, verify=False).text).get(
                             'return_code'))
        version = '{"name":"model version"}'
        self.assertEqual(
            json.loads(requests.post(HEADER + '/aiflow/metadata_store/modelVersion/get/version', data=version,
                                     verify=False).text).get(
                'return_code'), '2002')
        self.assertEqual('2002', json.loads(
            requests.post(HEADER + '/aiflow/metadata_store/job/get/id', data=data, verify=False).text).get(
            'return_code'))

    def test_delete_project_by_name(self):
        project = {
            "project": {"name": "project1"}
        }
        model = {
            "model": {"name": "model1",
                      "projectId": 2}
        }
        workflow_execution = {
            "workflowExecution": {"name": "execution1",
                                  "projectId": 2}
        }
        model_version = {
            "modelVersion": {"version": "model version1",
                             "modelId": 2,
                             "workflowExecutionId": 2}
        }
        job = {
            "job": {"name": "job1",
                    "workflowExecutionId": 2}
        }
        requests.post(HEADER + '/aiflow/metadata_store/project/save', data=json.dumps(project), verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/model/save', data=json.dumps(model), verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/save', data=json.dumps(workflow_execution),
                      verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/modelVersion/save', data=json.dumps(model_version), verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/job/save', data=json.dumps(job), verify=False)

        data = '{"name":"project1"}'
        requests.post(HEADER + '/aiflow/metadata_store/project/delete/name', data=data, verify=False)

        data = '{"id":2}'
        self.assertEqual('2002',
                         json.loads(requests.post(HEADER + '/aiflow/metadata_store/project/get/id', data=data,
                                                  verify=False).text).get('return_code'))
        self.assertEqual('2002',
                         json.loads(requests.post(HEADER + '/aiflow/metadata_store/model/get/id', data=data,
                                                  verify=False).text).get('return_code'))
        self.assertEqual('2002',
                         json.loads(requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/get/id',
                                                  data=data, verify=False).text).get(
                             'return_code'))
        version = '{"name":"model version"}'
        self.assertEqual(
            json.loads(requests.post(HEADER + '/aiflow/metadata_store/modelVersion/get/version', data=version,
                                     verify=False).text).get(
                'return_code'), '2002')
        self.assertEqual('2002', json.loads(
            requests.post(HEADER + '/aiflow/metadata_store/job/get/id', data=data, verify=False).text).get(
            'return_code'))

    def test_delete_model_by_id(self):
        model = {
            "model": {"name": "model_deleted",
                      "projectId": 1}
        }
        model_version = {
            "modelVersion": {"version": "model version deleted",
                             "modelId": 1,
                             "workflowExecutionId": 1}
        }
        requests.post(HEADER + '/aiflow/metadata_store/model/save', data=json.dumps(model), verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/modelVersion/save', data=json.dumps(model_version), verify=False)

        data = '{"id":1}'
        requests.post(HEADER + '/aiflow/metadata_store/model/delete/id', data=data, verify=False)

        model_name = '{"name": "model_deleted"}'
        self.assertEqual('2002', json.loads(
            requests.post(HEADER + '/aiflow/metadata_store/model/get/name', data=model_name, verify=False).text).get(
            'return_code'))

        version = '{"name":"model version deleted"}'
        self.assertEqual('2002', json.loads(
            requests.post(HEADER + '/aiflow/metadata_store/modelVersion/get/version', data=version,
                          verify=False).text).get(
            'return_code'))

    def test_delete_model_by_name(self):
        model = {
            "model": {"name": "model_deleted_name",
                      "projectId": 1}
        }
        model_version = {
            "modelVersion": {"version": "model version deleted name",
                             "modelId": 1,
                             "workflowExecutionId": 1}
        }
        requests.post(HEADER + '/aiflow/metadata_store/model/save', data=json.dumps(model), verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/modelVersion/save', data=json.dumps(model_version), verify=False)

        data = '{"name":"model_deleted_name"}'
        requests.post(HEADER + '/aiflow/metadata_store/model/delete/name', data=data, verify=False)

        model_name = '{"name": "model_deleted_name"}'
        self.assertEqual('2002', json.loads(
            requests.post(HEADER + '/aiflow/metadata_store/model/get/name', data=model_name, verify=False).text).get(
            'return_code'))

        version = '{"name":"model version deleted name"}'
        response = requests.post(HEADER + '/aiflow/metadata_store/modelVersion/get/version', data=version,
                                 verify=False)
        self.assertEqual('2002', json.loads(
            requests.post(HEADER + '/aiflow/metadata_store/modelVersion/get/version', data=version,
                          verify=False).text).get(
            'return_code'))

    def test_delete_workflow_execution_by_id(self):
        workflow_execution = {
            "workflowExecution": {"name": "execution_deleted_id",
                                  "projectId": 1}
        }
        model_version = {
            "modelVersion": {"version": "model version 1",
                             "modelId": 1,
                             "workflowExecutionId": 1}
        }
        job = {
            "job": {"name": "job",
                    "workflowExecutionId": 1}
        }
        requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/save', data=json.dumps(workflow_execution),
                      verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/modelVersion/save', data=json.dumps(model_version), verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/job/save', data=json.dumps(job), verify=False)
        data = '{"id":1}'
        requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/delete/id', data=data, verify=False)
        self.assertEqual('2002',
                         json.loads(requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/get/id',
                                                  data=data, verify=False).text).get(
                             'return_code'))
        version = '{"name":"model version 1"}'
        self.assertEqual('2002',
                         json.loads(
                             requests.post(HEADER + '/aiflow/metadata_store/modelVersion/get/version', data=version,
                                           verify=False).text).get(
                             'return_code'))
        self.assertEqual('2002',
                         json.loads(requests.post(HEADER + '/aiflow/metadata_store/job/get/id', data=data,
                                                  verify=False).text).get('return_code'))

    def test_delete_workflow_execution_by_name(self):
        workflow_execution = {
            "workflowExecution": {"name": "execution_deleted_name",
                                  "projectId": 1}
        }
        model_version = {
            "modelVersion": {"version": "model version 2",
                             "modelId": 1,
                             "workflowExecutionId": 1}
        }
        job = {
            "job": {"name": "job",
                    "workflowExecutionId": 1}
        }
        requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/save', data=json.dumps(workflow_execution),
                      verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/modelVersion/save', data=json.dumps(model_version), verify=False)
        requests.post(HEADER + '/aiflow/metadata_store/job/save', data=json.dumps(job), verify=False)
        data = '{"name": "execution_deleted_name"}'
        requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/delete/name', data=data, verify=False)
        self.assertEqual('2002',
                         json.loads(requests.post(HEADER + '/aiflow/metadata_store/workflowExecution/get/id',
                                                  data=data, verify=False).text).get(
                             'return_code'))
        version = '{"name":"model version 1"}'
        self.assertEqual('2002',
                         json.loads(
                             requests.post(HEADER + '/aiflow/metadata_store/modelVersion/get/version', data=version,
                                           verify=False).text).get(
                             'return_code'))
        self.assertEqual('2002',
                         json.loads(requests.post(HEADER + '/aiflow/metadata_store/job/get/id', data=data,
                                                  verify=False).text).get('return_code'))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from sqlalchemy import func

from airflow.api_connexion import security
from airflow.api_connexion.exceptions import NotFound
from airflow.api_connexion.parameters import check_limit, format_parameters
from airflow.api_connexion.schemas.event_log_schema import (
    EventLogCollection,
    event_log_collection_schema,
    event_log_schema,
)
from airflow.models import Log
from airflow.security import permissions
from airflow.utils.session import provide_session


@security.requires_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_AUDIT_LOG)])
@provide_session
def get_event_log(event_log_id, session):
    """Get a log entry"""
    event_log = session.query(Log).filter(Log.id == event_log_id).one_or_none()
    if event_log is None:
        raise NotFound("Event Log not found")
    return event_log_schema.dump(event_log)


@security.requires_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_AUDIT_LOG)])
@format_parameters({'limit': check_limit})
@provide_session
def get_event_logs(session, limit, offset=None):
    """Get all log entries from event log"""
    total_entries = session.query(func.count(Log.id)).scalar()
    event_logs = session.query(Log).order_by(Log.id).offset(offset).limit(limit).all()
    return event_log_collection_schema.dump(
        EventLogCollection(event_logs=event_logs, total_entries=total_entries)
    )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import pprint
from abc import abstractmethod


class Entity(object):

    def __iter__(self):
        # Iterate through list of properties and yield as key -> value
        for prop in self._properties():
            yield prop, self.__getattribute__(prop)

    @classmethod
    def _get_properties_helper(cls):
        return sorted([p for p in cls.__dict__ if isinstance(getattr(cls, p), property)])

    @classmethod
    def _properties(cls):
        return cls._get_properties_helper()

    @classmethod
    @abstractmethod
    def from_proto(cls, proto):
        pass

    def __repr__(self):
        return to_string(self)


def to_string(obj):
    return EntityPrinter().to_string(obj)


def get_classname(obj):
    return type(obj).__name__


class EntityPrinter(object):

    def __init__(self):
        super(EntityPrinter, self).__init__()
        self.printer = pprint.PrettyPrinter()

    def to_string(self, obj):
        if isinstance(obj, Entity):
            return "<%s: %s>" % (get_classname(obj), self._entity_to_string(obj))
        return self.printer.pformat(obj)

    def _entity_to_string(self, entity):
        return ", ".join(["%s=%s" % (key, self.to_string(value)) for key, value in entity])



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Serialized DAG table in database."""

import hashlib
import logging
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

import sqlalchemy_jsonfield
from sqlalchemy import BigInteger, Column, Index, String, and_
from sqlalchemy.orm import Session, backref, foreign, relationship
from sqlalchemy.sql import exists

from airflow.models.base import ID_LEN, Base
from airflow.models.dag import DAG, DagModel, DagEventDependencies
from airflow.models.dagcode import DagCode
from airflow.models.dagrun import DagRun
from airflow.serialization.serialized_objects import SerializedDAG
from airflow.settings import MIN_SERIALIZED_DAG_UPDATE_INTERVAL, json
from airflow.utils import timezone
from airflow.utils.session import provide_session
from airflow.utils.sqlalchemy import UtcDateTime

log = logging.getLogger(__name__)


class SerializedDagModel(Base):
    """A table for serialized DAGs.

    serialized_dag table is a snapshot of DAG files synchronized by scheduler.
    This feature is controlled by:

    * ``[core] min_serialized_dag_update_interval = 30`` (s):
      serialized DAGs are updated in DB when a file gets processed by scheduler,
      to reduce DB write rate, there is a minimal interval of updating serialized DAGs.
    * ``[scheduler] dag_dir_list_interval = 300`` (s):
      interval of deleting serialized DAGs in DB when the files are deleted, suggest
      to use a smaller interval such as 60

    It is used by webserver to load dags
    because reading from database is lightweight compared to importing from files,
    it solves the webserver scalability issue.
    """

    __tablename__ = 'serialized_dag'

    dag_id = Column(String(ID_LEN), primary_key=True)
    fileloc = Column(String(2000), nullable=False)
    # The max length of fileloc exceeds the limit of indexing.
    fileloc_hash = Column(BigInteger, nullable=False)
    data = Column(sqlalchemy_jsonfield.JSONField(json=json), nullable=False)
    # the event relationships of the dag
    event_relationships = Column(String(10240))
    last_updated = Column(UtcDateTime, nullable=False)
    dag_hash = Column(String(32), nullable=False)

    __table_args__ = (Index('idx_fileloc_hash', fileloc_hash, unique=False),)

    dag_runs = relationship(
        DagRun,
        primaryjoin=dag_id == foreign(DagRun.dag_id),
        backref=backref('serialized_dag', uselist=False, innerjoin=True),
    )

    dag_model = relationship(
        DagModel,
        primaryjoin=dag_id == DagModel.dag_id,  # type: ignore
        foreign_keys=dag_id,
        uselist=False,
        innerjoin=True,
        backref=backref('serialized_dag', uselist=False, innerjoin=True),
    )

    def __init__(self, dag: DAG):
        self.dag_id = dag.dag_id
        self.fileloc = dag.full_filepath
        self.fileloc_hash = DagCode.dag_fileloc_hash(self.fileloc)
        self.data = SerializedDAG.to_dict(dag)
        self.last_updated = timezone.utcnow()
        dag_event_deps = DagEventDependencies(dag)
        self.event_relationships = DagEventDependencies.to_json(dag_event_deps)
        self.dag_hash = hashlib.md5(json.dumps(self.data, sort_keys=True).encode("utf-8")).hexdigest()

    def __repr__(self):
        return f"<SerializedDag: {self.dag_id}>"

    @classmethod
    @provide_session
    def write_dag(cls, dag: DAG, min_update_interval: Optional[int] = None, session: Session = None):
        """Serializes a DAG and writes it into database.
        If the record already exists, it checks if the Serialized DAG changed or not. If it is
        changed, it updates the record, ignores otherwise.

        :param dag: a DAG to be written into database
        :param min_update_interval: minimal interval in seconds to update serialized DAG
        :param session: ORM Session
        """
        # Checks if (Current Time - Time when the DAG was written to DB) < min_update_interval
        # If Yes, does nothing
        # If No or the DAG does not exists, updates / writes Serialized DAG to DB
        if min_update_interval is not None:
            if session.query(
                exists().where(
                    and_(
                        cls.dag_id == dag.dag_id,
                        (timezone.utcnow() - timedelta(seconds=min_update_interval)) < cls.last_updated,
                    )
                )
            ).scalar():
                return

        log.debug("Checking if DAG (%s) changed", dag.dag_id)
        new_serialized_dag = cls(dag)
        serialized_dag_hash_from_db = session.query(cls.dag_hash).filter(cls.dag_id == dag.dag_id).scalar()

        if serialized_dag_hash_from_db == new_serialized_dag.dag_hash:
            log.debug("Serialized DAG (%s) is unchanged. Skipping writing to DB", dag.dag_id)
            return

        log.debug("Writing Serialized DAG: %s to the DB", dag.dag_id)
        session.merge(new_serialized_dag)
        log.debug("DAG: %s written to the DB", dag.dag_id)

    @classmethod
    @provide_session
    def read_all_dags(cls, session: Session = None) -> Dict[str, 'SerializedDAG']:
        """Reads all DAGs in serialized_dag table.

        :param session: ORM Session
        :returns: a dict of DAGs read from database
        """
        serialized_dags = session.query(cls)

        dags = {}
        for row in serialized_dags:
            log.debug("Deserializing DAG: %s", row.dag_id)
            dag = row.dag

            # Sanity check.
            if dag.dag_id == row.dag_id:
                dags[row.dag_id] = dag
            else:
                log.warning(
                    "dag_id Mismatch in DB: Row with dag_id '%s' has Serialised DAG with '%s' dag_id",
                    row.dag_id,
                    dag.dag_id,
                )
        return dags

    @property
    def dag(self):
        """The DAG deserialized from the ``data`` column"""
        if isinstance(self.data, dict):
            dag = SerializedDAG.from_dict(self.data)  # type: Any
        else:
            dag = SerializedDAG.from_json(self.data)  # noqa
        return dag

    @classmethod
    @provide_session
    def remove_dag(cls, dag_id: str, session: Session = None):
        """Deletes a DAG with given dag_id.

        :param dag_id: dag_id to be deleted
        :param session: ORM Session
        """
        # pylint: disable=no-member
        session.execute(cls.__table__.delete().where(cls.dag_id == dag_id))

    @classmethod
    @provide_session
    def remove_deleted_dags(cls, alive_dag_filelocs: List[str], session=None):
        """Deletes DAGs not included in alive_dag_filelocs.

        :param alive_dag_filelocs: file paths of alive DAGs
        :param session: ORM Session
        """
        alive_fileloc_hashes = [DagCode.dag_fileloc_hash(fileloc) for fileloc in alive_dag_filelocs]

        log.debug(
            "Deleting Serialized DAGs (for which DAG files are deleted) from %s table ", cls.__tablename__
        )

        # pylint: disable=no-member
        session.execute(
            cls.__table__.delete().where(
                and_(cls.fileloc_hash.notin_(alive_fileloc_hashes), cls.fileloc.notin_(alive_dag_filelocs))
            )
        )

    @classmethod
    @provide_session
    def has_dag(cls, dag_id: str, session: Session = None) -> bool:
        """Checks a DAG exist in serialized_dag table.

        :param dag_id: the DAG to check
        :param session: ORM Session
        """
        return session.query(exists().where(cls.dag_id == dag_id)).scalar()

    @classmethod
    @provide_session
    def get(cls, dag_id: str, session: Session = None) -> Optional['SerializedDagModel']:
        """
        Get the SerializedDAG for the given dag ID.
        It will cope with being passed the ID of a subdag by looking up the
        root dag_id from the DAG table.

        :param dag_id: the DAG to fetch
        :param session: ORM Session
        """
        row = session.query(cls).filter(cls.dag_id == dag_id).one_or_none()
        if row:
            return row

        # If we didn't find a matching DAG id then ask the DAG table to find
        # out the root dag
        root_dag_id = session.query(DagModel.root_dag_id).filter(DagModel.dag_id == dag_id).scalar()

        return session.query(cls).filter(cls.dag_id == root_dag_id).one_or_none()

    @staticmethod
    @provide_session
    def bulk_sync_to_db(dags: List[DAG], session: Session = None):
        """
        Saves DAGs as Serialized DAG objects in the database. Each
        DAG is saved in a separate database query.

        :param dags: the DAG objects to save to the DB
        :type dags: List[airflow.models.dag.DAG]
        :param session: ORM Session
        :type session: Session
        :return: None
        """
        for dag in dags:
            if not dag.is_subdag:
                SerializedDagModel.write_dag(
                    dag, min_update_interval=MIN_SERIALIZED_DAG_UPDATE_INTERVAL, session=session
                )

    @classmethod
    @provide_session
    def get_last_updated_datetime(cls, dag_id: str, session: Session = None) -> datetime:
        """
        Get the date when the Serialized DAG associated to DAG was last updated
        in serialized_dag table

        :param dag_id: DAG ID
        :type dag_id: str
        :param session: ORM Session
        :type session: Session
        """
        return session.query(cls.last_updated).filter(cls.dag_id == dag_id).scalar()

    @classmethod
    @provide_session
    def get_latest_version_hash(cls, dag_id: str, session: Session = None) -> str:
        """
        Get the latest DAG version for a given DAG ID.

        :param dag_id: DAG ID
        :type dag_id: str
        :param session: ORM Session
        :type session: Session
        :return: DAG Hash
        :rtype: str
        """
        return session.query(cls.dag_hash).filter(cls.dag_id == dag_id).scalar()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import List

from ai_flow.endpoint.server import stringValue, int64Value
from ai_flow.protobuf.message_pb2 import WorkflowProto, WorkflowExecutionProto, StateProto, JobProto
from ai_flow.plugin_interface.scheduler_interface import WorkflowInfo, WorkflowExecutionInfo, JobExecutionInfo
from ai_flow.workflow.status import Status


def proto_to_state(state):
    if state == StateProto.INIT:
        return Status.INIT
    elif state == StateProto.RUNNING:
        return Status.RUNNING
    elif state == StateProto.FINISHED:
        return Status.FINISHED
    elif state == StateProto.FAILED:
        return Status.FAILED
    elif state == StateProto.KILLED:
        return Status.KILLED


def workflow_to_proto(workflow: WorkflowInfo) -> WorkflowProto:
    if workflow is None:
        return None
    wp = WorkflowProto(name=workflow.workflow_name, namespace=workflow.namespace)
    for k, v in workflow.properties.items():
        wp.properties[k] = v
    return wp


def proto_to_workflow(proto: WorkflowProto) -> WorkflowInfo:
    if proto is None:
        return None
    else:
        return WorkflowInfo(namespace=proto.namespace, workflow_name=proto.name, properties=dict(proto.properties))


def workflow_list_to_proto(workflow_list: List[WorkflowInfo]) -> List[WorkflowProto]:
    result = []
    for workflow in workflow_list:
        result.append(workflow_to_proto(workflow))
    return result


def proto_to_workflow_list(proto_list: List[WorkflowProto]) -> List[WorkflowInfo]:
    result = []
    for proto in proto_list:
        result.append(proto_to_workflow(proto))
    return result


def workflow_execution_to_proto(workflow_execution: WorkflowExecutionInfo) -> WorkflowExecutionProto:
    if workflow_execution is None:
        return None
    if workflow_execution.status is None:
        state = Status.INIT
    else:
        state = workflow_execution.status
    if workflow_execution.start_date is None:
        start_date = '0'
    else:
        start_date = workflow_execution.start_date

    if workflow_execution.end_date is None:
        end_date = '0'
    else:
        end_date = workflow_execution.end_date

    wp = WorkflowExecutionProto(execution_id=workflow_execution.workflow_execution_id,
                                execution_state=StateProto.Value(state),
                                start_time=int64Value(int(start_date)),
                                end_time=int64Value(int(end_date)),
                                workflow=workflow_to_proto(workflow_execution.workflow_info))
    for k, v in workflow_execution.properties.items():
        wp.properties[k] = v
    return wp


def proto_to_workflow_execution(proto: WorkflowExecutionProto) -> WorkflowExecutionInfo:
    if proto is None:
        return None
    else:
        return WorkflowExecutionInfo(workflow_execution_id=proto.execution_id,
                                     status=proto_to_state(proto.execution_state),
                                     start_date=str(proto.start_time.value),
                                     end_date=str(proto.end_time.value),
                                     workflow_info=proto_to_workflow(proto.workflow),
                                     properties=dict(proto.properties))


def workflow_execution_list_to_proto(workflow_execution_list: List[WorkflowExecutionInfo]) \
        -> List[WorkflowExecutionProto]:
    result = []
    for workflow_execution in workflow_execution_list:
        result.append(workflow_execution_to_proto(workflow_execution))
    return result


def proto_to_workflow_execution_list(proto_list: List[WorkflowExecutionProto]) -> List[WorkflowExecutionInfo]:
    result = []
    for proto in proto_list:
        result.append(proto_to_workflow_execution(proto))
    return result


def job_to_proto(job: JobExecutionInfo) -> JobProto:
    if job.status is None:
        state = Status.INIT
    else:
        state = job.status

    if job.start_date is None:
        start_date = '0'
    else:
        start_date = job.start_date

    if job.end_date is None:
        end_date = '0'
    else:
        end_date = job.end_date

    return JobProto(name=job.job_name,
                    job_id=stringValue(job.job_execution_id),
                    job_state=StateProto.Value(state),
                    start_time=int64Value(int(start_date)),
                    end_time=int64Value(int(end_date)),
                    properties=dict(job.properties),
                    workflow_execution=workflow_execution_to_proto(job.workflow_execution))


def proto_to_job(proto: JobProto) -> JobExecutionInfo:
    if proto is None:
        return None
    else:
        return JobExecutionInfo(job_name=proto.name,
                                job_execution_id=proto.job_id.value,
                                status=proto_to_state(proto.job_state),
                                start_date=str(proto.start_time.value),
                                end_date=str(proto.end_time.value),
                                properties=dict(proto.properties),
                                workflow_execution=proto_to_workflow_execution(proto.workflow_execution))


def job_list_to_proto(job_list: List[JobExecutionInfo]) -> List[JobProto]:
    result = []
    for job in job_list:
        result.append(job_to_proto(job))
    return result


def proto_to_job_list(proto_list: List[JobProto]) -> List[JobExecutionInfo]:
    result = []
    for proto in proto_list:
        result.append(proto_to_job(proto))
    return result



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Remove id column from xcom

Revision ID: bbf4a7ad0465
Revises: cf5dc11e79ad
Create Date: 2019-10-29 13:53:09.445943

"""

from alembic import op
from sqlalchemy import Column, Integer
from sqlalchemy.engine.reflection import Inspector

# revision identifiers, used by Alembic.
revision = 'bbf4a7ad0465'
down_revision = 'cf5dc11e79ad'
branch_labels = None
depends_on = None


def upgrade():
    """Apply Remove id column from xcom"""
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)

    with op.batch_alter_table('xcom') as bop:
        xcom_columns = [col.get('name') for col in inspector.get_columns("xcom")]
        if "id" in xcom_columns:
            bop.drop_column('id')
            bop.drop_index('idx_xcom_dag_task_date')
            bop.create_primary_key('pk_xcom', ['dag_id', 'task_id', 'key', 'execution_date'])


def downgrade():
    """Unapply Remove id column from xcom"""
    with op.batch_alter_table('xcom') as bop:
        bop.drop_constraint('pk_xcom', type_='primary')
        bop.add_column(Column('id', Integer, primary_key=True))
        bop.create_index('idx_xcom_dag_task_date', ['dag_id', 'task_id', 'key', 'execution_date'])



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This is an example dag for using the Kubernetes Executor.
"""
import os

from airflow import DAG
from airflow.example_dags.libs.helper import print_stuff
from airflow.operators.python import PythonOperator
from airflow.utils.dates import days_ago

args = {
    'owner': 'airflow',
}

with DAG(
    dag_id='example_kubernetes_executor',
    default_args=args,
    schedule_interval=None,
    start_date=days_ago(2),
    tags=['example', 'example2'],
) as dag:

    affinity = {
        'podAntiAffinity': {
            'requiredDuringSchedulingIgnoredDuringExecution': [
                {
                    'topologyKey': 'kubernetes.io/hostname',
                    'labelSelector': {
                        'matchExpressions': [{'key': 'app', 'operator': 'In', 'values': ['airflow']}]
                    },
                }
            ]
        }
    }

    tolerations = [{'key': 'dedicated', 'operator': 'Equal', 'value': 'airflow'}]

    def use_zip_binary():
        """
        Checks whether Zip is installed.

        :return: True if it is installed, False if not.
        :rtype: bool
        """
        return_code = os.system("zip")
        if return_code != 0:
            raise SystemError("The zip binary is missing")

    # You don't have to use any special KubernetesExecutor configuration if you don't want to
    start_task = PythonOperator(task_id="start_task", python_callable=print_stuff)

    # But you can if you want to
    one_task = PythonOperator(
        task_id="one_task",
        python_callable=print_stuff,
        executor_config={"KubernetesExecutor": {"image": "airflow/ci:latest"}},
    )

    # Use the zip binary, which is only found in this special docker image
    two_task = PythonOperator(
        task_id="two_task",
        python_callable=use_zip_binary,
        executor_config={"KubernetesExecutor": {"image": "airflow/ci_zip:latest"}},
    )

    # Limit resources on this operator/task with node affinity & tolerations
    three_task = PythonOperator(
        task_id="three_task",
        python_callable=print_stuff,
        executor_config={
            "KubernetesExecutor": {
                "request_memory": "128Mi",
                "limit_memory": "128Mi",
                "tolerations": tolerations,
                "affinity": affinity,
            }
        },
    )

    # Add arbitrary labels to worker pods
    four_task = PythonOperator(
        task_id="four_task",
        python_callable=print_stuff,
        executor_config={"KubernetesExecutor": {"labels": {"foo": "bar"}}},
    )

    start_task >> [one_task, two_task, three_task, four_task]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from ai_flow.model_center.entity._model_repo_entity import _ModelRepoEntity
from ai_flow.protobuf.message_pb2 import ModelVersionStage


class ModelVersionParam(_ModelRepoEntity):
    """
    AIFlow entity for Model Version Parameter.
    """

    def __init__(self, model_path, model_type, version_desc, current_stage):
        self._model_path = model_path
        self._model_type = model_type
        self._version_desc = version_desc
        self._current_stage = current_stage

    @property
    def model_path(self):
        """String. Model source path for the model."""
        return self._model_path

    @property
    def model_type(self):
        """String. Model type of the model."""
        return self._model_type

    @property
    def version_desc(self):
        """String. Description of the model version."""
        return self._version_desc

    @property
    def current_stage(self):
        """String. Stage of the model version"""
        return self._current_stage

    # proto mappers
    @classmethod
    def from_proto(cls, proto):
        model_version = proto.model_version
        return cls(model_version.model_path.value if model_version.HasField("model_path") else None,
                   model_version.model_type.value if model_version.HasField("model_type") else None,
                   model_version.version_desc.value if model_version.HasField("version_desc") else None,
                   ModelVersionStage.Name(model_version.current_stage))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import asyncio
import logging

import grpc
import threading
import time
import uuid
from abc import ABC, abstractmethod

from typing import Dict

from ai_flow.protobuf.high_availability_pb2 import ListMembersResponse, ReturnStatus, MemberProto, \
    NotifyNewMemberResponse, NotifyNewMemberRequest
from ai_flow.protobuf.high_availability_pb2_grpc import HighAvailabilityManagerServicer, \
    HighAvailabilityManagerStub

if not hasattr(time, 'time_ns'):
    time.time_ns = lambda: int(time.time() * 1e9)


class Member(object):

    def __init__(self, version, server_uri, update_time):
        self.version = version
        self.server_uri = server_uri
        self.update_time = update_time


def member_to_proto(member: Member):
    return MemberProto(
        version=member.version, server_uri=member.server_uri, update_time=member.update_time)


def proto_to_member(member_proto):
    return Member(member_proto.version, member_proto.server_uri, member_proto.update_time)


def sleep_and_detecting_running(interval_ms, is_running_callable, min_interval_ms=500):
    start_time = time.time_ns() / 1000000
    while is_running_callable() and time.time_ns() / 1000000 < start_time + interval_ms:
        remaining = time.time_ns() / 1000000 - start_time
        if remaining > min_interval_ms:
            time.sleep(min_interval_ms / 1000)
        else:
            time.sleep(remaining / 1000)


class AIFlowServerHaManager(ABC):

    @abstractmethod
    def start(self, server_uri, storage, ttl_ms):
        pass

    @abstractmethod
    def get_living_members(self):
        pass

    @abstractmethod
    def add_living_member(self, member: Member):
        pass

    @abstractmethod
    def stop(self):
        pass


class SimpleAIFlowServerHaManager(AIFlowServerHaManager):

    def __init__(self):
        self.server_uri = None
        self.storage = None
        self.ttl_ms = None
        self.living_members = []
        self.member_connections = {}  # type: Dict[str, HighAvailabilityManagerStub]
        self.running = False
        self.heartbeat_thread = None
        self.notified_others_after_start = False
        self.uuid = str(uuid.uuid4())

    def start(self, server_uri, storage, ttl_ms):
        self.server_uri = server_uri
        self.storage = storage
        self.ttl_ms = ttl_ms
        self.running = True
        self.heartbeat_thread = threading.Thread(target=self.start_heartbeat, daemon=True)
        self.heartbeat_thread.start()

    def start_heartbeat(self):
        while self.running:
            try:
                # do heartbeat
                self.storage.clear_dead_members(self.ttl_ms)
                self.storage.update_member(self.server_uri, self.uuid)
                self.living_members = self.storage.list_living_members(self.ttl_ms)
                if not self.notified_others_after_start:
                    for member in self.living_members:
                        if member.server_uri == self.server_uri:
                            continue
                        channel = grpc.insecure_channel(member.server_uri)
                        self.member_connections[member.server_uri] = \
                            HighAvailabilityManagerStub(channel)
                        try:
                            self.member_connections[member.server_uri].notifyNewMember(
                                NotifyNewMemberRequest(member=member_to_proto(
                                    Member(1, self.server_uri, int(time.time_ns() / 1000000))
                                )))
                        except grpc.RpcError:
                            logging.error("Notify new member to '%s' failed." % member.server_uri,
                                          exc_info=True)
                    self.notified_others_after_start = True
            except Exception:
                logging.error("Exception thrown when send heartbeat to the HA storage.",
                              exc_info=True)
            sleep_and_detecting_running(self.ttl_ms / 2, lambda: self.running)

    def get_living_members(self):
        if len(self.living_members) == 0:
            self.living_members = self.storage.list_living_members(self.ttl_ms)
        return self.living_members

    def add_living_member(self, member: Member):
        if member.server_uri not in [member.server_uri for member in self.living_members]:
            self.living_members.append(member)

    def stop(self):
        self.running = False
        self.heartbeat_thread.join()


class HighAvailableService(HighAvailabilityManagerServicer):

    def __init__(self,
                 ha_manager,
                 server_uri,
                 ha_storage,
                 ttl_ms: int = 10000):
        self.ha_manager = ha_manager
        self.server_uri = server_uri
        self.ha_storage = ha_storage
        self.ttl_ms = ttl_ms
        self.member_updated_condition = asyncio.Condition()

    def start(self):
        self.ha_manager.start(self.server_uri,
                              self.ha_storage,
                              self.ttl_ms)
        logging.info("High available service started")

    def stop(self):
        self.ha_manager.stop()
        logging.info("High available service stopped")


    @asyncio.coroutine
    def listMembers(self, request, context):
        try:
            return self._list_members(request)
        except Exception as e:
            return ListMembersResponse(
                return_code=ReturnStatus.CALL_ERROR, return_msg=str(e))

    async def _list_members(self, request):
        timeout_seconds = request.timeout_seconds
        if 0 == timeout_seconds:
            return ListMembersResponse(
                return_code=ReturnStatus.CALL_SUCCESS,
                return_msg='',
                members=[member_to_proto(member) for member in self.ha_manager.get_living_members()])
        else:
            start = time.time()
            members = self.ha_manager.get_living_members()
            async with self.member_updated_condition:
                while time.time() - start < timeout_seconds:
                    try:
                        await asyncio.wait_for(self.member_updated_condition.wait(),
                                               timeout_seconds - time.time() + start)
                        members = self.ha_manager.get_living_members()
                        break
                    except asyncio.TimeoutError:
                        pass
            return ListMembersResponse(
                return_code=ReturnStatus.CALL_SUCCESS,
                return_msg='',
                members=[member_to_proto(member) for member in members])

    @asyncio.coroutine
    def notifyNewMember(self, request, context):
        try:
            return self._notify_new_member(request)
        except Exception as e:
            return ListMembersResponse(
                return_code=ReturnStatus.CALL_ERROR, return_msg=str(e))

    async def _notify_new_member(self, request):
        self.ha_manager.add_living_member(proto_to_member(request.member))
        async with self.member_updated_condition:
            self.member_updated_condition.notify_all()
        return NotifyNewMemberResponse(
            return_code=ReturnStatus.CALL_SUCCESS,
            return_msg='')



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import logging
import time

from ai_flow.meta.workflow_meta import WorkflowMeta
from ai_flow.store import MONGO_DB_ALIAS_META_SERVICE
from typing import Optional, Text, List, Union

import mongoengine

from ai_flow.common.status import Status
from ai_flow.meta.artifact_meta import ArtifactMeta
from ai_flow.meta.dataset_meta import DatasetMeta, Properties, DataType, Schema
from ai_flow.meta.metric_meta import MetricMeta, MetricType, MetricSummary
from ai_flow.meta.model_relation_meta import ModelRelationMeta, ModelVersionRelationMeta
from ai_flow.meta.project_meta import ProjectMeta
from ai_flow.metadata_store.utils.MetaToTable import MetaToTable
from ai_flow.metadata_store.utils.ResultToMeta import ResultToMeta
from ai_flow.metric.utils import table_to_metric_meta, table_to_metric_summary, metric_meta_to_table, \
    metric_summary_to_table
from ai_flow.model_center.entity.model_version_stage import STAGE_DELETED, get_canonical_stage, STAGE_GENERATED, \
    STAGE_DEPLOYED, STAGE_VALIDATED
from ai_flow.protobuf.message_pb2 import INVALID_PARAMETER_VALUE, RESOURCE_ALREADY_EXISTS
from ai_flow.endpoint.server.exception import AIFlowException
from ai_flow.endpoint.server.high_availability import Member
from ai_flow.store.abstract_store import AbstractStore
from ai_flow.store.db.db_model import (MongoProject, MongoDataset, MongoModelVersion,
                                       MongoArtifact, MongoRegisteredModel, MongoModelRelation,
                                       MongoMetricSummary, MongoMetricMeta,
                                       MongoModelVersionRelation, MongoMember, MongoProjectSnapshot, MongoWorkflow)
from ai_flow.store.db.db_util import parse_mongo_uri

if not hasattr(time, 'time_ns'):
    time.time_ns = lambda: int(time.time() * 1e9)

_logger = logging.getLogger(__name__)

TRUE = True
UPDATE_FAIL = -1
deleted_character = '~~'


class MongoStore(AbstractStore):
    """
    MongoDB backend store for tracking metadata for AIFlow backend entities.
    """

    CREATE_RETRY_TIMES = 3

    def __init__(self, **kwargs):
        """
        Create database backend storage by specified database URI

        :param db_uri: The mongodb database URI string to connect to the database.
        """
        super(MongoStore, self).__init__()
        db_conf = {
            "host": kwargs.get("host"),
            "port": kwargs.get("port"),
            "username": kwargs.get("username"),
            "password": kwargs.get("password"),
            "db": kwargs.get("db"),
        }
        self.db_uri = "mongodb://{username}:{password}@{host}:{port}/{db}?authSource=admin".format(**db_conf)
        MongoStoreConnManager().connect(self.db_uri, kwargs.get("db"))

    def _save_all(self, objs):
        """
        Backend entities store into database
        """
        if type(objs) is list:
            for obj in objs:
                obj.save()
        else:
            objs.save()

    """dataset api"""

    def get_dataset_by_id(self, dataset_id) -> Optional[DatasetMeta]:
        """
        get a specific dataset in metadata store by dataset id.

        :param dataset_id: the dataset id
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object if the dataset exists,
                 Otherwise, returns None if the dataset does not exist.
        """
        dataset_result = MongoDataset.objects(uuid=dataset_id, is_deleted__ne=TRUE)
        if len(dataset_result) == 0:
            return None
        return ResultToMeta.result_to_dataset_meta(dataset_result[0])

    def get_dataset_by_name(self, dataset_name) -> Optional[DatasetMeta]:
        """
        get a specific dataset in metadata store by dataset name.

        :param dataset_name: the dataset name
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object if the dataset exists,,
                 Otherwise, returns None if the dataset does not exist.
        """
        dataset_result = MongoDataset.objects(name=dataset_name, is_deleted__ne=TRUE)
        if len(dataset_result) == 0:
            return None
        return ResultToMeta.result_to_dataset_meta(dataset_result[0])

    def _register_dataset(self, name: Text, data_format: Text = None,
                          description: Text = None, uri: Text = None,
                          properties: Properties = None,
                          name_list: List[Text] = None, type_list: List[DataType] = None,
                          catalog_name: Text = None, catalog_type: Text = None,
                          catalog_connection_uri: Text = None,
                          catalog_table: Text = None, catalog_database: Text = None) -> DatasetMeta:
        """
        register an dataset in metadata store.

        :param name: the name of the dataset
        :param data_format: the data format of the dataset
        :param description: the description of the dataset
        :param uri: the uri of the dataset
        :param properties: the properties of the dataset
        :param name_list: the name list of dataset's schema
        :param type_list: the type list corresponded to the name list of dataset's schema
        :param catalog_name: the catalog name that will register in environment
        :param catalog_type: the catalog type of the dataset
        :param catalog_connection_uri: the connection uri of the catalog
        :param catalog_table: the table where the dataset is stored in the catalog
               if dataset is stored in the external catalog
        :param catalog_database: the database where the dataset is stored in the catalog
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object.
        """
        update_time = create_time = int(time.time() * 1000)
        before_dataset = self.get_dataset_by_name(dataset_name=name)
        if before_dataset is not None:
            # if the user has registered exactly the same dataset before,
            # do nothing in metadata store and return the registered dataset.
            if _compare_dataset_fields(data_format, description, uri,
                                       properties, name_list, type_list,
                                       catalog_name, catalog_type, catalog_database,
                                       catalog_connection_uri, catalog_table, before_dataset):
                return before_dataset
            else:
                # if the dataset registered this time has the same name but different fields,
                # raise the AIFlowException.
                raise AIFlowException("You have registered the dataset with same name: \"{}\" "
                                      "but different fields".format(name))
        try:
            if name_list is not None and type_list is not None:
                if len(name_list) != len(type_list):
                    raise AIFlowException("the length of name list and type list should be the same")
            if name_list is not None and type_list is None:
                raise AIFlowException("the length of name list and type list should be the same")
            if name_list is None and type_list is not None:
                raise AIFlowException("the length of name list and type list should be the same")
            dataset = MetaToTable.dataset_meta_to_table(name=name, data_format=data_format,
                                                        description=description, uri=uri, create_time=create_time,
                                                        update_time=update_time, properties=properties,
                                                        name_list=name_list, type_list=type_list,
                                                        catalog_name=catalog_name, catalog_type=catalog_type,
                                                        catalog_database=catalog_database,
                                                        catalog_connection_uri=catalog_connection_uri,
                                                        catalog_table=catalog_table,
                                                        store_type=type(self).__name__)
            dataset.save()
            schema = Schema(name_list=name_list, type_list=type_list)
            return DatasetMeta(uuid=dataset.uuid, name=name, data_format=data_format,
                               description=description, uri=uri, create_time=create_time, update_time=update_time,
                               properties=properties, schema=schema, catalog_name=catalog_name,
                               catalog_type=catalog_type, catalog_database=catalog_database,
                               catalog_connection_uri=catalog_connection_uri, catalog_table=catalog_table)
        except mongoengine.OperationError as e:
            raise AIFlowException('Registered Dataset (name={}) already exists. '
                                  'Error: {}'.format(dataset.name, str(e)))

    def register_dataset(self, name: Text, data_format: Text = None,
                         description: Text = None, uri: Text = None, properties: Properties = None,
                         name_list: List[Text] = None, type_list: List[DataType] = None):
        """
        register an dataset in metadata store.

        :param name: the name of the dataset
        :param data_format: the data format of the dataset
        :param description: the description of the dataset
        :param uri: the uri of the dataset
        :param properties: the properties of the dataset
        :param name_list: the name list of dataset's schema
        :param type_list: the type list corresponded to the name list of dataset's schema
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object.
        """
        return self._register_dataset(name=name,
                                      data_format=data_format, description=description, uri=uri,
                                      properties=properties, name_list=name_list, type_list=type_list)

    def register_dataset_with_catalog(self, name: Text, catalog_name: Text,
                                      catalog_type: Text, catalog_connection_uri: Text,
                                      catalog_table: Text, catalog_database: Text = None):
        """
        register an dataset in metadata store with catalog.

        :param name: the name of the dataset
        :param catalog_name: the catalog name that will register in environment
        :param catalog_type: the catalog type of the dataset
        :param catalog_connection_uri: the connection uri of the catalog
        :param catalog_table: the table where the dataset is stored in the catalog
        :param catalog_database: the database where the dataset is stored in the catalog
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object.
        """
        return self._register_dataset(name=name, catalog_name=catalog_name,
                                      catalog_type=catalog_type,
                                      catalog_connection_uri=catalog_connection_uri,
                                      catalog_table=catalog_table,
                                      catalog_database=catalog_database)

    def register_datasets(self, dataset_meta_list: List[DatasetMeta]) -> List[DatasetMeta]:
        """
        register multiple datasets in metadata store.

        :param dataset_meta_list: A list of datasets
        :return: List of :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` objects.
        """
        try:
            datasets = MetaToTable.dataset_meta_list_to_table(dataset_meta_list, store_type=type(self).__name__)
            for ex in datasets:
                ex.save()
            for dataset_meta, dataset in zip(dataset_meta_list, datasets):
                dataset_meta.uuid = dataset.uuid
            return dataset_meta_list
        except mongoengine.OperationError as e:
            raise AIFlowException(str(e))

    def update_dataset(self, dataset_name: Text,
                       data_format: Text = None, description: Text = None,
                       uri: Text = None, properties: Properties = None,
                       name_list: List[Text] = None, type_list: List[DataType] = None, catalog_name: Text = None,
                       catalog_type: Text = None, catalog_database: Text = None,
                       catalog_connection_uri: Text = None,
                       catalog_table: Text = None) -> Optional[DatasetMeta]:

        try:
            dataset: MongoDataset = MongoDataset.objects(name=dataset_name).first()
            if dataset is None:
                return None
            if data_format is not None:
                dataset.format = data_format
            if description is not None:
                dataset.description = description
            if uri is not None:
                dataset.uri = uri
            if properties is not None:
                dataset.properties = str(properties)
            if name_list is not None:
                dataset.name_list = str(name_list)
            if type_list is not None:
                data_type_list = []
                for data_type in type_list:
                    data_type_list.append(data_type.value)
                data_type_list = str(data_type_list)
                dataset.type_list = data_type_list
            if catalog_name is not None:
                dataset.catalog_name = catalog_name
            if catalog_type is not None:
                dataset.catalog_type = catalog_type
            if catalog_database is not None:
                dataset.catalog_database = catalog_database
            if catalog_connection_uri is not None:
                dataset.connection_config = catalog_connection_uri
            if catalog_table is not None:
                dataset.catalog_table = catalog_table
            dataset.update_time = int(time.time() * 1000)
            dataset.save()
            return ResultToMeta.result_to_dataset_meta(dataset)
        except mongoengine.OperationError as e:
            raise AIFlowException(e)

    def list_datasets(self, page_size, offset) -> Optional[List[DatasetMeta]]:
        """
        List registered datasets in metadata store.

        :param page_size: the limitation of the listed datasets.
        :param offset: the offset of listed datasets.
        :return: List of :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` objects,
                 return None if no datasets to be listed.
        """
        dataset_result = MongoDataset.objects(is_deleted__ne=TRUE).skip(offset).limit(page_size)
        if len(dataset_result) == 0:
            return None
        dataset_list = []
        for dataset in dataset_result:
            dataset_list.append(ResultToMeta.result_to_dataset_meta(dataset))
        return dataset_list

    def delete_dataset_by_name(self, dataset_name) -> Status:
        """
        Delete the registered dataset by dataset name .

        :param dataset_name: the dataset name
        :return: Status.OK if the dataset is successfully deleted, Status.ERROR if the dataset does not exist otherwise.
        """
        try:
            dataset = MongoDataset.objects(name=dataset_name, is_deleted__ne=TRUE).first()
            if dataset is None:
                return Status.ERROR
            deleted_dataset_counts = MongoDataset.objects(
                name__startswith=deleted_character + dataset_name + deleted_character, is_deleted=TRUE).count()
            dataset.is_deleted = TRUE
            dataset.name = deleted_character + dataset.name + deleted_character + str(deleted_dataset_counts + 1)
            dataset.save()
            return Status.OK
        except mongoengine.OperationError as e:
            raise AIFlowException(str(e))

    def delete_dataset_by_id(self, dataset_id) -> Status:
        """
        Delete the registered dataset by dataset id .

        :param dataset_id: the dataset id
        :return: Status.OK if the dataset is successfully deleted, Status.ERROR if the dataset does not exist otherwise.
        """
        dataset = self.get_dataset_by_id(dataset_id=dataset_id)
        if dataset is None:
            return Status.ERROR
        return self.delete_dataset_by_name(dataset.name)

    """project api"""

    def get_project_by_id(self, project_id) -> Optional[ProjectMeta]:
        """
        get a specific project in metadata store by project id

        :param project_id: the project id
        :return: A single :py:class:`ai_flow.meta.project.ProjectMeta` object if the project exists,
                 Otherwise, returns None if the project does not exist.
        """
        project_result = MongoProject.objects(uuid=project_id, is_deleted__ne=TRUE)
        if len(project_result) == 0:
            return None
        return ResultToMeta.result_to_project_meta(project_result[0])

    def get_project_by_name(self, project_name) -> Optional[ProjectMeta]:
        """
        get a specific project in metadata store by project name
        :param project_name: the project name
        :return: A single :py:class:`ai_flow.meta.project.ProjectMeta` object if the project exists,
                 Otherwise, returns None if the project does not exist.
        """
        project_result = MongoProject.objects(name=project_name, is_deleted__ne=TRUE)
        if len(project_result) == 0:
            return None
        return ResultToMeta.result_to_project_meta(project_result[0])

    def register_project(self, name: Text, uri: Text, properties: Properties = None) -> ProjectMeta:
        """
        register a project in metadata store.

        :param name: the name of the project
        :param uri: the uri of the project
        :param properties: the properties of the project
        :return: A single :py:class:`ai_flow.meta.project.ProjectMeta` object.
        """
        before_project = self.get_project_by_name(project_name=name)
        if before_project is not None:
            # if the user has registered exactly the same project before,
            # do nothing in metadata store and return the registered project.
            if _compare_project_fields(uri, properties, before_project):
                return before_project
            else:
                # if the project registered this time has the same name but different fields,
                # raise the AIFlowException.
                raise AIFlowException("You have registered the project with same name: \"{}\" "
                                      "but different fields".format(name))

        try:
            project = MetaToTable.project_meta_to_table(name=name, uri=uri, properties=properties,
                                                        store_type=type(self).__name__)
            project.save()
            project_meta = ProjectMeta(uuid=project.uuid, name=name, uri=uri, properties=properties)
            return project_meta
        except mongoengine.OperationError as e:
            raise AIFlowException('Registered Project (name={}) already exists. '
                                  'Error: {}'.format(project.name, str(e)))

    def list_project(self, page_size, offset) -> Optional[List[ProjectMeta]]:
        """
        List registered projects in metadata store.

        :param page_size: the limitation of the listed projects.
        :param offset: the offset of listed projects.
        :return: List of :py:class:`ai_flow.meta.project_meta.ProjectMeta` objects,
                 return None if no projects to be listed.
        """

        project_result = MongoProject.objects(is_deleted__ne=TRUE).skip(offset).limit(page_size)
        if len(project_result) == 0:
            return None
        projects = []
        for project in project_result:
            projects.append(ResultToMeta.result_to_project_meta(project))
        return projects

    def update_project(self, project_name: Text, uri: Text = None, properties: Properties = None) -> Optional[ProjectMeta]:
        try:
            project: MongoProject = MongoProject.objects(name=project_name, is_deleted__ne=TRUE).first()
            if project is None:
                return None
            if uri is not None:
                project.uri = uri
            if properties is not None:
                project.properties = str(properties)
            project.save()
            return ResultToMeta.result_to_project_meta(project)
        except mongoengine.OperationError as e:
            raise AIFlowException(e)

    def delete_project_by_id(self, project_id) -> Status:
        """
        Delete the registered project by project id .

        :param project_id: the project id
        :return: Status.OK if the project is successfully deleted, Status.ERROR if the project does not exist otherwise.
        """
        project = self.get_project_by_id(project_id=project_id)
        if project is None:
            return Status.ERROR
        return self.delete_project_by_name(project_name=project.name)

    def delete_project_by_name(self, project_name) -> Status:
        """
        Delete the registered project by project name .

        :param project_name: the project name
        :return: Status.OK if the project is successfully deleted, Status.ERROR if the project does not exist otherwise.
        """
        try:
            project = MongoProject.objects(name=project_name, is_deleted__ne=TRUE).first()
            if project is None:
                return Status.ERROR
            deleted_project_counts = MongoProject.objects(
                name__startswith=deleted_character + project_name + deleted_character,
                is_deleted=TRUE).count()
            project.is_deleted = TRUE
            project.name = deleted_character + project.name + deleted_character + str(deleted_project_counts + 1)
            model_version_list = []
            for per_model in project.model_relation:
                deleted_model_relation_counts = MongoModelRelation.objects(
                    name__startswith=deleted_character + per_model.name + deleted_character,
                    is_deleted=TRUE).count()
                per_model.is_deleted = TRUE
                per_model.name = deleted_character + per_model.name + deleted_character + str(
                    deleted_model_relation_counts + 1)
                for model_version in per_model.model_version_relation:
                    deleted_model_version_relation_counts = MongoModelVersionRelation.objects(
                        version__startswith=deleted_character + model_version.version + deleted_character,
                        is_deleted=TRUE).count()
                    model_version.is_deleted = TRUE
                    model_version.version = deleted_character + model_version.version + deleted_character + str(
                        deleted_model_version_relation_counts + 1)
                    # update for unqine constraint
                    model_version.version_model_id_unique = f'{model_version.version}-{model_version.model_id}'
                    model_version.version_workflow_execution_id_unique = f'{model_version.version}-{model_version.workflow_execution_id}'
                model_version_list += per_model.model_version_relation
            self._save_all(
                [project] + project.model_relation + model_version_list)
            return Status.OK
        except mongoengine.OperationError as e:
            raise AIFlowException(str(e))

    '''workflow api'''

    def register_workflow(self, name, project_id, properties=None) -> WorkflowMeta:
        """
        Register a workflow in metadata store.

        :param name: the workflow name
        :param project_id: the id of project which contains the workflow
        :param properties: the workflow properties
        """
        update_time = create_time = int(time.time() * 1000)
        try:
            workflow = MetaToTable.workflow_to_table(name=name,
                                                     project_id=project_id,
                                                     properties=properties,
                                                     create_time=create_time,
                                                     update_time=update_time,
                                                     store_type=type(self).__name__)
            workflow.save()
            return WorkflowMeta(uuid=workflow.uuid, name=name,
                                project_id=project_id, properties=properties,
                                create_time=create_time, update_time=update_time)
        except mongoengine.OperationError as e:
            raise AIFlowException('Registered Workflow (name={}, project_id={}) already exists. '
                                  'Error: {}'.format(workflow.name, workflow.project_id, str(e)))

    def get_workflow_by_name(self, project_name, workflow_name) -> Optional[WorkflowMeta]:
        """
        Get a workflow by specific project name and workflow name

        :param project_name: the name of project which contains the workflow
        :param workflow_name: the workflow name
        """
        project = self.get_project_by_name(project_name)
        if not project:
            return None
        workflow_result = MongoWorkflow.objects(name=workflow_name, project_id=project.uuid, is_deleted__ne=TRUE)
        if len(workflow_result) == 0:
            return None
        return ResultToMeta.result_to_workflow_meta(workflow_result[0])

    def get_workflow_by_id(self, workflow_id) -> Optional[WorkflowMeta]:
        """
        Get a workflow by specific uuid

        :param workflow_id: the uuid of workflow
        """
        workflow_result = MongoWorkflow.objects(uuid=workflow_id, is_deleted__ne=TRUE)
        if len(workflow_result) == 0:
            return None
        return ResultToMeta.result_to_workflow_meta(workflow_result[0])

    def list_workflows(self, project_name, page_size, offset) -> Optional[List[WorkflowMeta]]:
        """
        List all workflows of the specific project

        :param project_name: the name of project which contains the workflow
        :param page_size     limitation of listed workflows.
        :param offset        offset of listed workflows.
        """
        project = self.get_project_by_name(project_name)
        if not project:
            return None
        workflow_result = MongoWorkflow.objects(project_id=project.uuid, is_deleted__ne=TRUE)\
            .skip(offset).limit(page_size)
        if len(workflow_result) == 0:
            return None
        workflows = []
        for workflow in workflow_result:
            workflows.append(ResultToMeta.result_to_workflow_meta(workflow))
        return workflows

    def delete_workflow_by_name(self, project_name, workflow_name) -> Status:
        """
        Delete the workflow by specific project and workflow name

        :param project_name: the name of project which contains the workflow
        :param workflow_name: the workflow name
        """

        workflow = self.get_workflow_by_name(project_name=project_name,
                                             workflow_name=workflow_name)
        if workflow is None:
            return Status.ERROR
        else:
            return self.delete_workflow_by_id(workflow.uuid)

    def delete_workflow_by_id(self, workflow_id) -> Status:
        """
        Delete the workflow by specific id

        :param workflow_id: the uuid of workflow
        """
        try:
            workflow = MongoWorkflow.objects(uuid=workflow_id, is_deleted__ne=TRUE).first()
            if workflow is None:
                return Status.ERROR
            deleted_workflow_counts = MongoWorkflow.objects(
                project_id=workflow.project_id,
                name__startswith=deleted_character + workflow.name + deleted_character,
                is_deleted=TRUE).count()
            workflow.is_deleted = TRUE
            workflow.name = deleted_character + workflow.name + deleted_character + str(deleted_workflow_counts + 1)
            workflow.save()
            return Status.OK
        except mongoengine.OperationError as e:
            raise AIFlowException(str(e))

    def update_workflow(self, workflow_name, project_name, properties=None) -> Optional[WorkflowMeta]:
        """
        Update the workflow

        :param workflow_name: the workflow name
        :param project_name: the name of project which contains the workflow
        :param properties: (Optional) the properties need to be updated
        """
        try:
            project = self.get_project_by_name(project_name)
            if not project:
                raise AIFlowException("The project name you specific doesn't exists, project: \"{}\""
                                      .format(project_name))
            workflow: MongoWorkflow = MongoWorkflow.objects(name=workflow_name,
                                                            project_id=project.uuid,
                                                            is_deleted__ne=TRUE).first()
            if workflow is None:
                return None
            if properties is not None:
                workflow.properties = properties
            workflow.update_time = int(time.time() * 1000)
            workflow.save()
            return ResultToMeta.result_to_workflow_meta(workflow)
        except mongoengine.OperationError as e:
            raise AIFlowException(e)

    """model api"""

    def get_model_relation_by_id(self, model_id) -> Optional[ModelRelationMeta]:
        """
        get a specific model relation in metadata store by model id.

        :param model_id: the model id
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` object if the model relation
                 exists, Otherwise, returns None if the model relation does not exist.
        """
        model_result = MongoModelRelation.objects(uuid=model_id, is_deleted__ne=TRUE)
        if len(model_result) == 0:
            return None
        return ResultToMeta.result_to_model_relation_meta(model_result[0])

    def get_model_relation_by_name(self, name) -> Optional[ModelRelationMeta]:
        """
        get a specific model relation in metadata store by model name.

        :param name: the model name
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` object if the model relation
                 exists, Otherwise, returns None if the model relation does not exist.
        """
        model_result = MongoModelRelation.objects(name=name, is_deleted__ne=TRUE)
        if len(model_result) == 0:
            return None
        return ResultToMeta.result_to_model_relation_meta(model_result[0])

    def register_model_relation(self, name: Text,
                                project_id: int = None) -> ModelRelationMeta:
        """
        register a model relation in metadata store

        :param name: the name of the model
        :param project_id: the project id which the model corresponded to.
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` object.
        """
        before_model_relation = self.get_model_relation_by_name(name=name)
        if before_model_relation is not None:
            # if the user has registered exactly the same model relation before,
            # do nothing in metadata store and return the registered model relation.
            if _compare_model_relation_fields(project_id, before_model_relation):
                return before_model_relation
            else:
                # if the dataset registered this time has the same name but different fields,
                # raise the AIFlowException.
                raise AIFlowException("You have registered the model relation with same name: \"{}\" "
                                      "but different project uuid".format(name))
        try:
            model = MetaToTable.model_relation_meta_to_table(name=name,
                                                             project_id=project_id,
                                                             store_type=type(self).__name__)
            model.save()
            # update reference field
            if project_id is not None:
                project = MongoProject.objects(uuid=project_id).first()
                if project.model_relation is None:
                    project.model_relation = [model]
                else:
                    project.model_relation.append(model)
                project.save()
            model_meta = ModelRelationMeta(uuid=model.uuid, name=name, project_id=project_id)
            return model_meta
        except mongoengine.OperationError as e:
            raise AIFlowException('Registered Model (name={}) already exists. '
                                  'Error: {}'.format(model.name, str(e)))

    def list_model_relation(self, page_size, offset) -> Optional[List[ModelRelationMeta]]:
        """
        List registered model relations in metadata store.

        :param page_size: the limitation of the listed model relations.
        :param offset: the offset of listed model relations.
        :return: List of :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` objects,
                 return None if no model relations to be listed.
        """
        model_result = MongoModelRelation.objects(is_deleted__ne=TRUE).skip(offset).limit(page_size)
        if len(model_result) == 0:
            return None
        models = []
        for model in model_result:
            models.append(ResultToMeta.result_to_model_relation_meta(model))
        return models

    def delete_model_relation_by_id(self, model_id) -> Status:
        """
        Delete the registered model by model id .

        :param model_id: the model id
        :return: Status.OK if the model is successfully deleted, Status.ERROR if the model does not exist otherwise.
        """
        model = self.get_model_relation_by_id(model_id=model_id)
        if model is None:
            return Status.ERROR
        return self.delete_model_relation_by_name(model_name=model.name)

    def delete_model_relation_by_name(self, model_name) -> Status:
        """
        Delete the registered model by model name .

        :param model_name: the model name
        :return: Status.OK if the model is successfully deleted, Status.ERROR if the model does not exist otherwise.
        """
        try:
            model = MongoModelRelation.objects(name=model_name, is_deleted__ne=TRUE).first()
            if model is None:
                return Status.ERROR
            deleted_model_counts = MongoModelRelation.objects(
                name__startswith=deleted_character + model.name + deleted_character,
                is_deleted=TRUE).count()
            model.is_deleted = TRUE
            model.name = deleted_character + model.name + deleted_character + str(deleted_model_counts + 1)
            for model_version in model.model_version_relation:
                deleted_model_version_counts = MongoModelVersionRelation.objects(
                    version__startswith=deleted_character + model_version.version + deleted_character,
                    is_deleted=TRUE).count()
                model_version.is_deleted = TRUE
                model_version.version = deleted_character + model_version.version + deleted_character + str(
                    deleted_model_version_counts + 1)
                # update for unqine constraint
                model_version.version_model_id_unique = f'{model_version.version}-{model_version.model_id}'
                model_version.version_workflow_execution_id_unique = f'{model_version.version}-{model_version.workflow_execution_id}'
            self._save_all([model] + model.model_version_relation)
            return Status.OK
        except mongoengine.OperationError as e:
            raise AIFlowException(str(e))

    '''model version api'''

    def get_model_version_relation_by_version(self, version_name, model_id) -> Optional[ModelVersionRelationMeta]:
        """
        get a specific model version relation in metadata store by the model version name.

        :param version_name: the model version name
        :param model_id: the model id corresponded to the model version
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelVersionRelationMeta` object
                 if the model version exists, Otherwise, returns None if the model version does not exist.
        """
        model_version_result = MongoModelVersionRelation.objects(version=version_name, model_id=model_id,
                                                                 is_deleted__ne=TRUE)
        if len(model_version_result) == 0:
            return None
        return ResultToMeta.result_to_model_version_relation_meta(model_version_result[0])

    def register_model_version_relation(self, version: Text,
                                        model_id: int,
                                        project_snapshot_id: int = None) -> ModelVersionRelationMeta:
        """
        register a model version relation in metadata store.

        :param version: the specific model version
        :param model_id: the model id corresponded to the model version
        :param project_snapshot_id: the project snapshot id corresponded to the model version
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelVersionRelationMeta` object.
        """
        try:
            model_version_relation = MetaToTable.model_version_relation_to_table(version=version,
                                                                                 model_id=model_id,
                                                                                 project_snapshot_id=project_snapshot_id,
                                                                                 store_type=type(self).__name__)
            model_version_relation.save()
            # update reference field
            if model_id is not None:
                model_relation = MongoModelRelation.objects(uuid=model_id).first()
                if model_relation.model_version_relation is None:
                    model_relation.model_version_relation = [model_version_relation]
                else:
                    model_relation.model_version_relation.append(model_version_relation)
                model_relation.save()

            if project_snapshot_id is not None:
                project_snapshot = MongoProjectSnapshot.objects(uuid=project_snapshot_id).first()
                if project_snapshot.model_version_relation is None:
                    project_snapshot.model_version_relation = [model_version_relation]
                else:
                    project_snapshot.model_version_relation.append(model_version_relation)
                project_snapshot.save()

            model_version_relation_meta = ModelVersionRelationMeta(version=version, model_id=model_id,
                                                                   project_snapshot_id=project_snapshot_id)
            return model_version_relation_meta
        except mongoengine.OperationError as e:
            raise AIFlowException('Registered ModelVersion (name={}) already exists. '
                                  'Error: {}'.format(model_version_relation.version, str(e)))

    def list_model_version_relation(self, model_id, page_size, offset) -> Optional[List[ModelVersionRelationMeta]]:
        """
        List registered model version relations in metadata store.

        :param model_id: the model id corresponded to the model version
        :param page_size: the limitation of the listed model version relations.
        :param offset: the offset of listed model version relations.
        :return: List of :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` objects,
                 return None if no model version relations to be listed.
        """
        model_version_result = MongoModelVersionRelation.objects(model_id=model_id, is_deleted__ne=TRUE).skip(
            offset).limit(page_size)
        if len(model_version_result) == 0:
            return None
        model_version_list = []
        for version in model_version_result:
            model_version_list.append(ResultToMeta.result_to_model_version_relation_meta(version))
        return model_version_list

    def delete_model_version_relation_by_version(self, version, model_id) -> Status:
        """
        Delete the registered model version by model version name .

        :param version: the model version name
        :param model_id: the model id corresponded to the model version
        :return: Status.OK if the model version is successfully deleted,
                 Status.ERROR if the model version does not exist otherwise.
        """
        try:
            model_version = MongoModelVersionRelation.objects(version=version, model_id=model_id,
                                                              is_deleted__ne=TRUE).first()
            if model_version is None:
                return Status.ERROR
            deleted_model_version_counts = MongoModelVersionRelation.objects(
                version__startswith=deleted_character + version + deleted_character,
                is_deleted=TRUE).count()
            model_version.is_deleted = TRUE
            model_version.version = deleted_character + model_version.version + deleted_character + str(
                deleted_model_version_counts + 1)
            # update for unqine constraint
            model_version.version_model_id_unique = f'{model_version.version}-{model_version.model_id}'
            model_version.version_workflow_execution_id_unique = f'{model_version.version}-{model_version.workflow_execution_id}'
            model_version.save()
            return Status.OK
        except mongoengine.OperationError as e:
            raise AIFlowException(str(e))

    """artifact api"""

    def get_artifact_by_id(self, artifact_id: int) -> Optional[ArtifactMeta]:
        """
        get a specific artifact in metadata store by artifact id.

        :param artifact_id: the artifact id
        :return: A single :py:class:`ai_flow.meta.artifact_meta.ArtifactMeta` object
                 if the artifact exists, Otherwise, returns None if the artifact does not exist.
        """
        artifact_result = MongoArtifact.objects(uuid=artifact_id, is_deleted__ne=TRUE)
        if len(artifact_result) == 0:
            return None
        artifact = ResultToMeta.result_to_artifact_meta(artifact_result[0])
        return artifact

    def get_artifact_by_name(self, artifact_name: Text) -> Optional[ArtifactMeta]:
        """
        get a specific artifact in metadata store by artifact name.

        :param artifact_name: the artifact name
        :return: A single :py:class:`ai_flow.meta.artifact_meta.ArtifactMeta` object
                 if the artifact exists, Otherwise, returns None if the artifact does not exist.
        """
        artifact_result = MongoArtifact.objects(name=artifact_name, is_deleted__ne=TRUE)
        if len(artifact_result) == 0:
            return None
        artifact = ResultToMeta.result_to_artifact_meta(artifact_result[0])
        return artifact

    def register_artifact(self, name: Text, artifact_type: Text = None, description: Text = None,
                          uri: Text = None,
                          properties: Properties = None) -> ArtifactMeta:
        """
        register an artifact in metadata store.

        :param name: the name of the artifact
        :param artifact_type: the type of the artifact
        :param description: the description of the artifact
        :param uri: the uri of the artifact
        :param properties: the properties of the artifact
        :return: A single :py:class:`ai_flow.meta.artifact_meta.py.ArtifactMeta` object.
        """
        before_artifact = self.get_artifact_by_name(artifact_name=name)
        if before_artifact is not None:
            # if the user has registered exactly the same artifact before,
            # do nothing in metadata store and return the registered artifact.
            if _compare_artifact_fields(artifact_type, description, uri,
                                        properties, before_artifact):
                return before_artifact
            else:
                # if the artifact registered this time has the same name but different fields,
                # raise the AIFlowException.
                raise AIFlowException("You have registered the artifact with same name: \"{}\""
                                      " but different fields".format(name))
        try:
            create_time = update_time = int(time.time() * 1000)
            artifact = MetaToTable.artifact_meta_to_table(name=name, artifact_type=artifact_type,
                                                          description=description, uri=uri,
                                                          create_time=create_time,
                                                          update_time=update_time, properties=properties,
                                                          store_type=type(self).__name__)
            artifact.save()
            artifact_meta = ArtifactMeta(uuid=artifact.uuid, name=name, artifact_type=artifact_type,
                                         description=description, uri=uri, create_time=create_time,
                                         update_time=update_time, properties=properties)
            return artifact_meta
        except mongoengine.OperationError as e:
            raise AIFlowException('Registered Artifact (name={}) already exists. '
                                  'Error: {}'.format(artifact.name, str(e)))

    def update_artifact(self, name: Text, artifact_type: Text = None,
                        description: Text = None, uri: Text = None,
                        properties: Properties = None) -> Optional[ArtifactMeta]:
        try:
            artifact: MongoArtifact = MongoArtifact.objects(name=name, is_deleted__ne=TRUE).first()
            if artifact is None:
                return None
            if artifact_type is not None:
                artifact.artifact_type = artifact_type
            if description is not None:
                artifact.description = description
            if uri is not None:
                artifact.uri = uri
            if properties is not None:
                artifact.properties = str(properties)
            artifact.update_time = int(time.time() * 1000)
            artifact.save()
            return ResultToMeta.result_to_artifact_meta(artifact)
        except mongoengine.OperationError as e:
            raise AIFlowException(str(e))

    def list_artifact(self, page_size, offset) -> Optional[List[ArtifactMeta]]:
        """
        List registered artifacts in metadata store.

        :param page_size: the limitation of the listed artifacts.
        :param offset: the offset of listed artifacts.
        :return: List of :py:class:`ai_flow.meta.artifact_meta.py.ArtifactMeta` objects,
                 return None if no artifacts to be listed.
        """
        artifact_result = MongoArtifact.objects(is_deleted__ne=TRUE).skip(offset).limit(page_size)
        if len(artifact_result) == 0:
            return None
        artifact_list = []
        for artifact in artifact_result:
            artifact_list.append(ResultToMeta.result_to_artifact_meta(artifact))
        return artifact_list

    def delete_artifact_by_id(self, artifact_id) -> Status:
        """
        Delete the registered artifact by artifact id .

        :param artifact_id: the artifact id
        :return: Status.OK if the artifact is successfully deleted,
                 Status.ERROR if the artifact does not exist otherwise.
        """
        artifact = self.get_artifact_by_id(artifact_id=artifact_id)
        if artifact is None:
            return Status.ERROR
        return self.delete_artifact_by_name(artifact.name)

    def delete_artifact_by_name(self, artifact_name) -> Status:
        """
        Delete the registered artifact by artifact name .

        :param artifact_name: the artifact name
        :return: Status.OK if the artifact is successfully deleted,
                 Status.ERROR if the artifact does not exist otherwise.
        """
        try:
            artifact = MongoArtifact.objects(name=artifact_name, is_deleted__ne=TRUE).first()
            if artifact is None:
                return Status.ERROR
            deleted_artifact_counts = MongoArtifact.objects(
                name__startswith=deleted_character + artifact_name + deleted_character,
                is_deleted=TRUE).count()
            artifact.is_deleted = TRUE
            artifact.name = deleted_character + artifact.name + deleted_character + str(deleted_artifact_counts + 1)
            artifact.save()
            return Status.OK
        except mongoengine.OperationError as e:
            raise AIFlowException(str(e))

    def get_deployed_model_version(self, model_name):
        return self.get_model_with_stage(model_name, STAGE_DEPLOYED)

    def get_latest_validated_model_version(self, model_name):
        return self.get_model_with_stage(model_name, STAGE_VALIDATED)

    def get_latest_generated_model_version(self, model_name):
        return self.get_model_with_stage(model_name, STAGE_GENERATED)

    def get_model_with_stage(self, model_name, stage):
        if model_name is None:
            raise AIFlowException('Registered model name cannot be empty.', INVALID_PARAMETER_VALUE)
        model_version = MongoModelVersion.objects(model_name=model_name, current_stage=stage)
        if len(model_version) == 0:
            return None
        else:
            return model_version[len(model_version) - 1].to_meta_entity()

    @classmethod
    def _get_registered_model(cls, model_name):
        """
        Query registered model in Model Center filter by model name

        :param model_name: Unique name for registered model within Model Center.
        """
        if model_name is None:
            raise AIFlowException('Registered model name cannot be empty.', INVALID_PARAMETER_VALUE)

        register_models = MongoRegisteredModel.objects(model_name=model_name)

        if len(register_models) == 0:
            return None
        else:
            _logger.info("Get registered model name: %s, versions: %s.", register_models[0].model_name,
                         register_models[0].model_version)
            return register_models[0]

    def create_registered_model(self, model_name, model_desc=None):
        """
        Create a new registered model in model repository.

        :param model_name: Name of registered model. This is expected to be unique in the backend store.
        :param model_desc: (Optional) Description of registered model.

        :return: Object of :py:class:`ai_flow.model_center.entity.RegisteredModel` created in Model Center.
        """
        if model_name is None:
            raise AIFlowException('Registered model name cannot be empty.', INVALID_PARAMETER_VALUE)
        try:
            before_model = self._get_registered_model(model_name=model_name)
            if before_model is not None:
                if _compare_model_fields(model_desc, before_model):
                    registered_model = MongoRegisteredModel(model_name=model_name,
                                                            model_desc=model_desc)
                    return registered_model.to_meta_entity()
                else:
                    raise AIFlowException("You have registered the model with same name: \"{}\" "
                                          "but different fields".format(model_name), RESOURCE_ALREADY_EXISTS)
            registered_model = MongoRegisteredModel(model_name=model_name,
                                                    model_desc=model_desc)
            registered_model.save()
            return registered_model.to_meta_entity()
        except mongoengine.OperationError as e:
            raise AIFlowException('Registered Model (name={}) already exists. Error: {}'.format(model_name, str(e)),
                                  RESOURCE_ALREADY_EXISTS)

    def update_registered_model(self, registered_model, model_name=None, model_desc=None):
        """
        Update metadata for RegisteredModel entity. Either ``model_name`` or ``model_desc``
        should be non-None. Backend raises exception if registered model with given name does not exist.

        :param registered_model: :py:class:`ai_flow.model_center.entity.RegisteredModel` object.
        :param model_name: (Optional) New proposed name for the registered model.
        :param model_desc: (Optional) Description of registered model.

        :return: A single updated :py:class:`ai_flow.model_center.entity.RegisteredModel` object.
        """
        registered_model = self._get_registered_model(registered_model.model_name)
        if registered_model is None:
            return None
        else:
            try:
                if model_name is not None:
                    registered_model.model_name = model_name
                    # Update model name of registered model version
                    for model_version in registered_model.model_version:
                        model_version.model_name = model_name
                if model_desc is not None:
                    registered_model.model_desc = model_desc
                self._save_all([registered_model] + registered_model.model_version)
                return registered_model.to_meta_entity()
            except mongoengine.OperationError as e:
                raise AIFlowException(
                    'Registered model (name={}) already exists. Error: {}'.format(model_name, str(e)),
                    RESOURCE_ALREADY_EXISTS)

    def delete_registered_model(self, registered_model):
        """
        Delete registered model based on specific model name.
        Backend raises exception if registered model with given name does not exist.

        :param registered_model: :py:class:`ai_flow.model_center.entity.RegisteredModel` object.

        :return: None
        """
        try:
            registered_model = self._get_registered_model(registered_model.model_name)
            registered_model.delete()
            return Status.OK
        except mongoengine.OperationError as e:
            raise AIFlowException(str(e))

    def list_registered_models(self):
        """
        List of registered models backend in Model Center.

        :return: List of :py:class:`ai_flow.model_center.entity.RegisteredModel` objects.
        """
        return [registered_model.to_detail_entity() for registered_model in MongoRegisteredModel.objects()]

    def get_registered_model_detail(self, registered_model):
        """
        Get registered model detail filter by model name and model version for Model Center.

        :param registered_model: :py:class:`ai_flow.model_center.entity.RegisteredModel` object.

        :return: Object of :py:class:`ai_flow.model_center.entity.RegisteredModelDetail` backend in Model Center.
        """
        registered_model = self._get_registered_model(registered_model.model_name)
        return None if registered_model is None else registered_model.to_detail_entity()

    @classmethod
    def _get_model_version(cls, model_version):
        model_name = model_version.model_name
        model_version = model_version.model_version
        if model_name is None:
            raise AIFlowException('Registered model name cannot be empty.', INVALID_PARAMETER_VALUE)
        if model_version is None:
            raise AIFlowException('Registered model version cannot be empty.', INVALID_PARAMETER_VALUE)

        model_versions = MongoModelVersion.objects(model_name=model_name, model_version=model_version,
                                                   current_stage__ne=STAGE_DELETED)

        if len(model_versions) == 0:
            return None
        else:
            _logger.info("Get registered model version: %s of model name: %s.", model_versions[0],
                         model_versions[0].model_name)
            return model_versions[0]

    @classmethod
    def _list_model_versions(cls, registered_model):
        model_name = registered_model.model_name
        if model_name is None:
            raise AIFlowException('Registered model name cannot be empty.', INVALID_PARAMETER_VALUE)

        return MongoModelVersion.objects(model_name=model_name, current_stage__ne=STAGE_DELETED)

    @classmethod
    def _count_deleted_model_version(cls, model_version):
        model_name = model_version.model_name
        model_version = model_version.model_version
        if model_name is None:
            raise AIFlowException('Registered model name cannot be empty.', INVALID_PARAMETER_VALUE)
        if model_version is None:
            raise AIFlowException('Registered model version cannot be empty.', INVALID_PARAMETER_VALUE)

        return MongoModelVersion.objects(model_name=model_name,
                                         model_version__startswith=deleted_character + model_version + deleted_character,
                                         current_stage__ne=STAGE_DELETED).count()

    def create_model_version(self, model_name, model_path, model_type=None,
                             version_desc=None, current_stage=STAGE_GENERATED):
        """
        Create a new model version from given model source and model metric.

        :param model_name: Name for containing registered model.
        :param model_path: Source path where the AIFlow model is stored.
        :param model_type: (Optional) Type of AIFlow registered model option.
        :param version_desc: (Optional) Description of registered model version.
        :param current_stage: (Optional) Stage of registered model version

        :return: Object of :py:class:`ai_flow.model_center.entity.ModelVersion` created in Model Center.
        """

        def next_version(current_version):
            if current_version is None:
                return "1"
            else:
                return str(current_version + 1)

        for attempt in range(self.CREATE_RETRY_TIMES):
            try:
                registered_model = self._get_registered_model(model_name)
                if registered_model is None:
                    return None
                else:
                    model_versions = self._list_model_versions(registered_model)
                    if model_versions is None:
                        version_num = 0
                    else:
                        version_num = len(model_versions)
                    model_version = next_version(version_num)
                    doc_model_version = MongoModelVersion(model_name=model_name,
                                                          model_version=model_version,
                                                          model_path=model_path,
                                                          model_type=model_type,
                                                          version_desc=version_desc,
                                                          current_stage=get_canonical_stage(current_stage))
                    doc_model_version.save()
                    # update reference field
                    if model_name is not None:
                        registered_model = MongoRegisteredModel.objects(model_name=model_name).first()
                        if registered_model.model_version is None:
                            registered_model.model_version = [doc_model_version]
                        else:
                            registered_model.model_version.append(doc_model_version)
                        registered_model.save()
                    return doc_model_version.to_meta_entity()
            except mongoengine.OperationError as e:
                logging.info(model_version)
                more_retries = self.CREATE_RETRY_TIMES - attempt - 1
                _logger.info(
                    'Create model version (model_version=%s) error (model_name=%s). Retrying %s more time%s.',
                    model_version, model_name,
                    str(more_retries), 's' if more_retries > 1 else '')
        raise AIFlowException(
            'Create model version error (model_name={}). Giving up after {} attempts.'.format(model_name,
                                                                                              self.CREATE_RETRY_TIMES))

    def update_model_version(self, model_version, model_path=None, model_type=None,
                             version_desc=None, current_stage=None):
        """
        Update metadata associated with a model version in model repository.

        :param model_version: :py:class:`ai_flow.model_center.entity.ModelVersion` object.
        :param model_path: (Optional) New Source path where AIFlow model is stored.
        :param model_type: (Optional) Type of AIFlow registered model option.
        :param version_desc: (Optional) New Description of registered model version.
        :param current_stage: (Optional) New desired stage for this model version.

        :return: A single updated :py:class:`ai_flow.model_center.entity.ModelVersion` object.
        """
        serving_model_version = self.get_deployed_model_version(model_version.model_name)
        if serving_model_version is not None and current_stage == 'DEPLOYED':
            raise AIFlowException('There is already a serving model version="{}" of model="{}"'.
                                  format(serving_model_version.model_version, serving_model_version.model_name))
        model_version = self._get_model_version(model_version)
        if model_version is None:
            return None
        else:
            try:
                if model_path is not None:
                    model_version.model_path = model_path
                if model_type is not None:
                    model_version.model_type = model_type
                if version_desc is not None:
                    model_version.version_desc = version_desc
                if current_stage is not None:
                    model_version.current_stage = get_canonical_stage(current_stage)
                model_version.save()
                return model_version.to_meta_entity()
            except mongoengine.OperationError as e:
                raise AIFlowException(
                    'Update model version error (model_name={}, model_version={}).'.format(model_version.model_name,
                                                                                           model_version.model_version))

    def delete_model_version(self, model_version):
        """
        Delete model version in model repository.

        :param model_version: :py:class:`ai_flow.model_center.entity.ModelVersion` object.

        :return: None
        """
        doc_model_version = self._get_model_version(model_version)
        if doc_model_version is None:
            return None

        doc_model_version.model_path = "REDACTED-SOURCE-PATH"
        doc_model_version.model_type = "REDACTED-TYPE-FEATURE"
        doc_model_version.version_status = None
        doc_model_version.version_desc = None
        doc_model_version.current_stage = STAGE_DELETED
        doc_model_version.save()

    def get_model_version_detail(self, model_version):
        """
        :param model_version: :py:class:`ai_flow.model_center.entity.ModelVersion` object.

        :return: Object of :py:class:`ai_flow.model_center.entity.ModelVersionDetail` backend in Model Center.
        """
        doc_model_version = self._get_model_version(model_version)
        return None if doc_model_version is None else doc_model_version.to_meta_entity()

    """metric api"""

    def register_metric_meta(self, metric_name, metric_type, project_name, metric_desc=None, dataset_name=None,
                             model_name=None, job_name=None, start_time=None, end_time=None, uri=None, tags=None,
                             properties=None) -> MetricMeta:
        try:
            metric_meta_table = metric_meta_to_table(metric_name, metric_type, metric_desc, project_name,
                                                     dataset_name, model_name, job_name, start_time, end_time, uri,
                                                     tags, properties, type(self).__name__)
            metric_meta_table.save()
            return MetricMeta(metric_name=metric_name,
                              metric_type=metric_type,
                              metric_desc=metric_desc,
                              project_name=project_name,
                              dataset_name=dataset_name,
                              model_name=model_name,
                              job_name=job_name,
                              start_time=start_time,
                              end_time=end_time,
                              uri=uri,
                              tags=tags,
                              properties=properties)
        except Exception as e:
            raise AIFlowException('Register metric meta failed! Error: {}.'.format(str(e)))

    def update_metric_meta(self, metric_name, metric_desc=None, project_name=None, dataset_name=None,
                           model_name=None, job_name=None, start_time=None, end_time=None, uri=None, tags=None,
                           properties=None) -> MetricMeta:
        try:
            metric_meta_table: MongoMetricMeta = MongoMetricMeta.objects(metric_name=metric_name,
                                                                         is_deleted__ne=TRUE).first()
            if metric_desc is not None:
                metric_meta_table.metric_desc = metric_desc
            if project_name is not None:
                metric_meta_table.project_name = project_name
            if dataset_name is not None:
                metric_meta_table.dataset_name = dataset_name
            if model_name is not None:
                metric_meta_table.model_name = model_name
            if job_name is not None:
                metric_meta_table.job_name = job_name
            if start_time is not None:
                metric_meta_table.start_time = start_time
            if end_time is not None:
                metric_meta_table.end_time = end_time
            if uri is not None:
                metric_meta_table.uri = uri
            if tags is not None:
                metric_meta_table.tags = tags
            if properties is not None and properties != {}:
                metric_meta_table.properties = str(properties)
            metric_meta_table.save()
            return table_to_metric_meta(metric_meta_table)
        except Exception as e:
            raise AIFlowException('Update metric meta failed! Error: {}.'.format(str(e)))

    def delete_metric_meta(self, metric_name):
        try:
            metric_meta = MongoMetricMeta.objects(metric_name=metric_name, is_deleted__ne=TRUE).first()
            if metric_meta is None:
                return Status.ERROR
            deleted_metric_meta_counts = MongoMetricMeta.objects(
                name__startswith=deleted_character + metric_meta.metric_name + deleted_character,
                is_deleted=TRUE).count()
            metric_meta.is_deleted = TRUE
            metric_meta.metric_name = deleted_character + metric_meta.metric_name + deleted_character + str(
                deleted_metric_meta_counts + 1)
            for metric_summary in metric_meta.metric_summary:
                deleted_metric_summary_counts = MongoMetricSummary.objects(
                    name__startswith=deleted_character + metric_summary.metric_name + deleted_character,
                    is_deleted=TRUE).count()
                metric_summary.is_deleted = TRUE
                metric_summary.metric_name = deleted_character + metric_summary.metric_name + deleted_character + str(
                    deleted_metric_summary_counts + 1)
            self._save_all([metric_meta] + metric_meta.metric_summary)
            return Status.OK
        except Exception as e:
            raise AIFlowException('Delete metric meta failed! Error: {}.'.format(str(e)))

    def get_metric_meta(self, metric_name) -> Union[None, MetricMeta]:
        try:
            metric_meta_table = MongoMetricMeta.objects(metric_name=metric_name, is_deleted__ne=TRUE).first()
            if metric_meta_table is None:
                return None
            else:
                return table_to_metric_meta(metric_meta_table)
        except Exception as e:
            raise AIFlowException('Get metric meta failed! Error: {}.'.format(str(e)))

    def list_dataset_metric_metas(self, dataset_name, project_name=None) -> Union[None, MetricMeta, List[MetricMeta]]:
        try:
            if project_name is None:
                metric_meta_tables = MongoMetricMeta.objects(dataset_name=dataset_name,
                                                             metric_type=MetricType.DATASET.value,
                                                             is_deleted__ne=TRUE)
            else:
                metric_meta_tables = MongoMetricMeta.objects(dataset_name=dataset_name,
                                                             project_name=project_name,
                                                             metric_type=MetricType.DATASET.value,
                                                             is_deleted__ne=TRUE)
            if len(metric_meta_tables) == 0:
                return None
            elif len(metric_meta_tables) == 1:
                metric_meta_table = metric_meta_tables[0]
                return table_to_metric_meta(metric_meta_table)
            else:
                result = []
                for metric_meta_table in metric_meta_tables:
                    result.append(table_to_metric_meta(metric_meta_table))
                return result
        except Exception as e:
            raise AIFlowException('Get dataset metric metas failed! Error: {}.'.format(str(e)))

    def list_model_metric_metas(self, model_name, project_name=None) -> Union[
            None, MetricMeta, List[MetricMeta]]:
        try:
            if project_name is None:
                metric_meta_tables = MongoMetricMeta.objects(model_name=model_name,
                                                             metric_type=MetricType.MODEL.value,
                                                             is_deleted__ne=TRUE)
            else:
                metric_meta_tables = MongoMetricMeta.objects(model_name=model_name,
                                                             project_name=project_name,
                                                             metric_type=MetricType.MODEL.value,
                                                             is_deleted__ne=TRUE)
            if len(metric_meta_tables) == 0:
                return None
            elif len(metric_meta_tables) == 1:
                metric_meta_table = metric_meta_tables[0]
                return table_to_metric_meta(metric_meta_table)
            else:
                result = []
                for metric_meta_table in metric_meta_tables:
                    result.append(table_to_metric_meta(metric_meta_table))
                return result
        except Exception as e:
            raise AIFlowException('Get model metric metas failed! Error: {}.'.format(str(e)))

    def register_metric_summary(self, metric_name, metric_key, metric_value, metric_timestamp, model_version=None,
                                job_execution_id=None) -> MetricSummary:
        try:
            metric_summary_table = metric_summary_to_table(metric_name, metric_key, metric_value, metric_timestamp,
                                                           model_version, job_execution_id)
            metric_summary_table.save()
            return MetricSummary(uuid=metric_summary_table.uuid,
                                 metric_name=metric_name,
                                 metric_key=metric_key,
                                 metric_value=metric_value,
                                 metric_timestamp=metric_timestamp,
                                 model_version=model_version,
                                 job_execution_id=job_execution_id)
        except mongoengine.OperationError as e:
            raise AIFlowException('Register metric summary failed! Error: {}.'.format(str(e)))

    def update_metric_summary(self, uuid, metric_name=None, metric_key=None, metric_value=None, metric_timestamp=None,
                              model_version=None, job_execution_id=None) -> MetricSummary:
        try:
            metric_summary_table = MongoMetricSummary.objects(uuid=uuid, is_deleted__ne=TRUE).first()
            if metric_name is not None:
                metric_summary_table.metric_name = metric_name
            if metric_key is not None:
                metric_summary_table.metric_key = metric_key
            if metric_value is not None:
                metric_summary_table.metric_value = metric_value
            if metric_timestamp is not None:
                metric_summary_table.metric_timestamp = metric_timestamp
            if model_version is not None:
                metric_summary_table.model_version = model_version
            if job_execution_id is not None:
                metric_summary_table.job_execution_id = job_execution_id
            metric_summary_table.save()
            return table_to_metric_summary(metric_summary_table)
        except Exception as e:
            raise AIFlowException('Update metric summary failed! Error: {}.'.format(str(e)))

    def delete_metric_summary(self, uuid: int):
        try:
            metric_summary_table = MongoMetricSummary.objects(uuid=uuid).first()
            metric_summary_table.is_deleted = TRUE
            metric_summary_table.save()
        except Exception as e:
            raise AIFlowException('Delete metric summary failed! Error: {}.'.format(str(e)))

    def get_metric_summary(self, uuid) -> Union[None, MetricSummary]:
        try:
            metric_summary_table = MongoMetricSummary.objects(uuid=uuid,
                                                              is_deleted__ne=TRUE)
            if metric_summary_table is None:
                return None
            else:
                return table_to_metric_summary(metric_summary_table)
        except Exception as e:
            raise AIFlowException('Get metric summary failed! Error: {}.'.format(str(e)))

    def list_metric_summaries(self, metric_name=None, metric_key=None, model_version=None, start_time=None,
                              end_time=None) -> Union[None, MetricSummary, List[MetricSummary]]:
        try:
            metric_summary_tables = MongoMetricSummary.objects(is_deleted__ne=TRUE)
            if metric_name is not None:
                metric_summary_tables = MongoMetricSummary.objects(metric_name=metric_name,
                                                                   is_deleted__ne=TRUE)
            if metric_key is not None:
                metric_summary_tables = MongoMetricSummary.objects(metric_key=metric_key,
                                                                   is_deleted__ne=TRUE)
            if model_version is not None:
                metric_summary_tables = MongoMetricSummary.objects(model_version=model_version,
                                                                   is_deleted__ne=TRUE)
            if start_time is not None:
                metric_summary_tables = MongoMetricSummary.objects(start_time__ge=start_time,
                                                                   is_deleted__ne=TRUE)
            if end_time is not None:
                metric_summary_tables = MongoMetricSummary.objects(start_time__le=end_time,
                                                                   is_deleted__ne=TRUE)
            if len(metric_summary_tables) == 0:
                return None
            elif len(metric_summary_tables) == 1:
                metric_summary_table = metric_summary_tables[0]
                return table_to_metric_summary(metric_summary_table)
            else:
                results = []
                for metric_summary_table in metric_summary_tables:
                    results.append(table_to_metric_summary(metric_summary_table))
                return results
        except Exception as e:
            raise AIFlowException('List metric summaries failed! Error: {}.'.format(str(e)))

    """member api"""

    def list_living_members(self, ttl_ms) -> List[Member]:
        try:
            member_models = MongoMember.objects(update_time__gte=time.time_ns() / 1000000 - ttl_ms)
            return [Member(m.version, m.server_uri, int(m.update_time)) for m in member_models]
        except Exception as e:
            raise AIFlowException("List living AIFlow Member Error.") from e

    def update_member(self, server_uri, server_id):
        try:
            member = MongoMember.objects(server_uri=server_uri).first()
            if member is None:
                member = MongoMember()
                member.version = 1
                member.server_uri = server_uri
                member.update_time = time.time_ns() / 1000000
                member.uuid = server_id
                member.save()
            else:
                if member.uuid != server_id:
                    raise Exception("The server uri '%s' is already exists in the storage!" %
                                    server_uri)
                member.version += 1
                member.update_time = time.time_ns() / 1000000
                member.save()
        except Exception as e:
            raise AIFlowException("Update AIFlow Member Error.") from e

    def clear_dead_members(self, ttl_ms):
        try:
            members = MongoMember.objects(update_time__lt=time.time_ns() / 1000000 - ttl_ms)
            if len(members) == 0:
                return None
            else:
                for member in members:
                    member.delete()
        except Exception as e:
            raise AIFlowException("Clear dead AIFlow Member Error.") from e


def _compare_dataset_fields(data_format, description, uri,
                            properties, name_list, type_list, catalog_name,
                            catalog_type, catalog_database, catalog_connection_uri, catalog_table,
                            before_dataset) -> bool:
    return data_format == before_dataset.data_format and description == before_dataset.description \
           and uri == before_dataset.uri \
           and properties == before_dataset.properties and name_list == before_dataset.schema.name_list \
           and type_list == before_dataset.schema.type_list and catalog_name == before_dataset.catalog_name \
           and catalog_type == before_dataset.catalog_type and catalog_database == before_dataset.catalog_database \
           and catalog_connection_uri == before_dataset.catalog_connection_uri \
           and catalog_table == before_dataset.catalog_table


def _compare_artifact_fields(artifact_type, description, uri, properties, before_artifact):
    return artifact_type == before_artifact.artifact_type and description == before_artifact.description \
           and uri == before_artifact.uri and properties == before_artifact.properties


def _compare_model_fields(model_desc, before_model):
    return model_desc == before_model.model_desc


def _compare_model_relation_fields(project_id, before_model_relation):
    return project_id == before_model_relation.project_id


def _compare_project_fields(uri, properties, before_project):
    return uri == before_project.uri and properties == before_project.properties


class MongoStoreConnManager(object):
    _instance = None
    _conns = {}
    _connected_uris = set()
    _closed_uris = set()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = object.__new__(cls, *args, **kwargs)
        return cls._instance

    def connect(self, db_uri, db_name, db_alias=MONGO_DB_ALIAS_META_SERVICE):
        if db_uri not in self._conns:
            self._conns[db_uri] = mongoengine.connect(db_name, host=db_uri, alias=db_alias)
            self._connected_uris.add(db_uri)

    def disconnect(self, db_uri, db_alias=MONGO_DB_ALIAS_META_SERVICE):
        if db_uri in self._conns:
            mongoengine.disconnect(alias=db_alias)
            self._connected_uris.remove(db_uri)
            self._closed_uris.add(db_uri)

    def disconnect_all(self):
        current_connected_uris = self._connected_uris.copy()
        for uri in current_connected_uris:
            self.disconnect(uri)

    def drop(self, db_uri_without_auth):
        db_uri = f'{db_uri_without_auth}?authSource=admin'
        if db_uri in self._conns:
            _, _, _, _, db = parse_mongo_uri(db_uri_without_auth)
            self._conns[db_uri].drop_database(db)

    def drop_all(self):
        current_connected_uris = self._connected_uris.copy()
        for db_uri in current_connected_uris:
            self.drop(db_uri)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Example DAG demonstrating the DummyOperator and a custom DummySkipOperator which skips by default."""

from airflow import DAG
from airflow.exceptions import AirflowSkipException
from airflow.operators.dummy import DummyOperator
from airflow.utils.dates import days_ago

args = {
    'owner': 'airflow',
}


# Create some placeholder operators
class DummySkipOperator(DummyOperator):
    """Dummy operator which always skips the task."""

    ui_color = '#e8b7e4'

    def execute(self, context):
        raise AirflowSkipException


def create_test_pipeline(suffix, trigger_rule, dag_):
    """
    Instantiate a number of operators for the given DAG.

    :param str suffix: Suffix to append to the operator task_ids
    :param str trigger_rule: TriggerRule for the join task
    :param DAG dag_: The DAG to run the operators on
    """
    skip_operator = DummySkipOperator(task_id=f'skip_operator_{suffix}', dag=dag_)
    always_true = DummyOperator(task_id=f'always_true_{suffix}', dag=dag_)
    join = DummyOperator(task_id=trigger_rule, dag=dag_, trigger_rule=trigger_rule)
    final = DummyOperator(task_id=f'final_{suffix}', dag=dag_)

    skip_operator >> join
    always_true >> join
    join >> final


dag = DAG(dag_id='example_skip_dag', default_args=args, start_date=days_ago(2), tags=['example'])
create_test_pipeline('1', 'all_success', dag)
create_test_pipeline('2', 'one_success', dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Set


class WeightRule:
    """Weight rules."""

    DOWNSTREAM = 'downstream'
    UPSTREAM = 'upstream'
    ABSOLUTE = 'absolute'

    _ALL_WEIGHT_RULES = set()  # type: Set[str]

    @classmethod
    def is_valid(cls, weight_rule):
        """Check if weight rule is valid."""
        return weight_rule in cls.all_weight_rules()

    @classmethod
    def all_weight_rules(cls):
        """Returns all weight rules"""
        if not cls._ALL_WEIGHT_RULES:
            cls._ALL_WEIGHT_RULES = {
                getattr(cls, attr)
                for attr in dir(cls)
                if not attr.startswith("_") and not callable(getattr(cls, attr))
            }
        return cls._ALL_WEIGHT_RULES
